#include "expmodel.h"
#include <string.h>
#include <set>
#include <math.h>
#include <iostream>
#include <algorithm>
#include <exception>


using std::set;
using namespace std;

typedef unsigned int uint;



ExpModel::ExpModel(uint n, uint d, bool use_items, bool use_rowmargins) : item_cnt(n), trans_cnt(d), include_items(use_items), include_rowmargins(use_rowmargins)
{
	PartitionNode::set_n(item_cnt);
	PartitionNode::set_d(trans_cnt);
	PartitionNode::set_mode(include_items, include_rowmargins);
	mpfr_set_default_prec(max(PartitionNode::prob_prec, PartitionNode::size_prec));
	
	BitVector::set_size(item_cnt);
	
	blocks = new PartitionBlock[max_partition_size];
	inited_blocks = 0;
	partition_size = 1;
	
	blocks[0].index = 0;
	blocks[0].node = new PartitionNode();
	++inited_blocks;
	
	mpfr_set_ui(blocks[0].node->size, 1, GMP_RNDN);
	mpfr_set_ui(blocks[0].node->prob, 1, GMP_RNDN);
	if (include_items || include_rowmargins) {
		mpfr_set_ui(blocks[0].node->cum_size, 1, GMP_RNDN);
	}
	
	itemsets = new Itemset*[max_k];
	frequencies = new double[max_k];
	itemsets_bits = new BitVector*[max_k];
	for (uint i=0; i<max_k; ++i)
		itemsets_bits[i] = new BitVector();
	itemsets_count = 0;
	itemsets_union = new uint[item_cnt];
	itemsets_union_size = 0;
	itemsets_union_bits = new BitVector();
	
	dummy_itemset_adder = new Itemset(0, 0, false);
	
	item_frequencies = new double[item_cnt];
	for (uint i=0; i<item_cnt; ++i)
		item_frequencies[i] = 0.5;

	if (include_items || include_rowmargins) {
		item_params = new double[item_cnt];
		for (uint i=0; i<item_cnt; ++i)
			item_params[i] = 0.5;
	}
	else {
		item_params = 0;
	}
	
	if (include_rowmargins) {
		rowsize_probs = new double[item_cnt+1];
		w_params = new mpfr_t[item_cnt+1];
		rowsize_estimates = new mpfr_t[item_cnt+1];
		for (uint ii=0; ii<=item_cnt; ++ii) {
			mpfr_init_set_ui(w_params[ii], 1, GMP_RNDN);
			mpfr_init(rowsize_estimates[ii]);
		}
	}
	else {
		rowsize_probs = 0;
		w_params = 0;
		rowsize_estimates = 0;
	}
	
	mpfr_init_set_ui(u0_param, 1, GMP_RNDN);
	mpfr_init_set_ui(v0_param, 1, GMP_RNDN); // note: also used for uniform background model
	mpfr_init_set_ui(w0_param, 1, GMP_RNDN);
	if (include_rowmargins) {
		mpfr_mul_2si(u0_param, u0_param, -(int)item_cnt, GMP_RNDN);
		mpfr_mul_2si(v0_param, v0_param, -(int)item_cnt, GMP_RNDN);
	}
	
	u_params = new mpfr_t[max_k];
	for (uint i=0; i<max_k; ++i) {
		mpfr_init(u_params[i]);
	}
	
	if (include_items) {
		v_params = new mpfr_t[item_cnt];
		for (uint i=0; i<item_cnt; ++i) {
			mpfr_init_set_ui(v_params[i], 1, GMP_RNDN);
		}
	}
	else
		v_params = 0;
	
	bitvect = new BitVector();
	
	mpfr_init(estimate);
	mpfr_init(sum);
	mpfr_init(factor);
	mpfr_init(h);
	mpfr_init(f);
	mpfr_init(f1);
	mpfr_init(f2);
	mpfr_init(tmp);
	
	if (include_rowmargins) {
		compute_all_cum_sizes();
		for (uint ii=0; ii<=item_cnt; ++ii)
			mpfr_set(blocks[0].node->size_array[ii], blocks[0].node->cum_size_array[ii], GMP_RNDN);
	}
	
	return;
}


ExpModel::~ExpModel()
{
	delete [] blocks;
	
	for (uint i=0; i<itemsets_count; ++i)
		delete itemsets[i];
	delete [] itemsets;
	delete [] frequencies;
	for (uint i=0; i<max_k; ++i)
		delete itemsets_bits[i];
	delete [] itemsets_bits;
	delete [] itemsets_union;
	delete itemsets_union_bits;
	
	dummy_itemset_adder->itemset = 0;
	delete dummy_itemset_adder;
	
	delete [] item_frequencies;
	delete [] item_params;
	
	if (include_rowmargins) {
		for (uint ii=0; ii<=item_cnt; ++ii)
			mpfr_clear(rowsize_estimates[ii]);
	}
	delete [] rowsize_probs;
	delete [] rowsize_estimates;
	
	if (include_rowmargins) {
		for (uint ii=0; ii<=item_cnt; ++ii)
			mpfr_clear(w_params[ii]);
	}
	delete [] w_params;
	mpfr_clear(w0_param);
	
	mpfr_clear(u0_param);
	for (uint i=0; i<max_k; ++i) {
		mpfr_clear(u_params[i]);
	}
	
	mpfr_clear(v0_param);
	if (include_items) {
		for (uint i=0; i<item_cnt; ++i)
			mpfr_clear(v_params[i]);
	}
	
	delete [] u_params;
	delete [] v_params;
	
	delete bitvect;
	
	mpfr_clear(estimate);
	mpfr_clear(sum);
	mpfr_clear(factor);
	mpfr_clear(h);
	mpfr_clear(f);
	mpfr_clear(f1);
	mpfr_clear(f2);
	mpfr_clear(tmp);
	
	return;
}

// For include_rowmargins==true, we do already use all items


// these should be set before anything else is done
// @pre: item_frequencies != 0
// itemset_count should be 0, and partition_size should be 1
void ExpModel::set_item_frequency(int item, double frequency)
{
	item_frequencies[item] = frequency;
	if (include_items) {
		// TODO: initialize with independence distribution, not uniform
		if (!itemsets_union_bits->isset(item)) {
			if (include_rowmargins) {
				mpfr_mul_d(v0_param, v0_param, (1.0 - frequency) / (1.0 - item_params[item]), GMP_RNDN);
				mpfr_set_d(v_params[item], frequency / (1.0 - frequency), GMP_RNDN);
				mpfr_mul_d(u0_param, u0_param, (1.0 - frequency) / (1.0 - item_params[item]), GMP_RNDN);
			}
			else {
				// note: we're not adding the item to the distrubution here yet
				//mpfr_mul_d(v0_param, v0_param, (1.0 - frequency) , GMP_RNDN);
				mpfr_set_d(v_params[item], frequency / (1.0 - frequency), GMP_RNDN);
				//mpfr_mul_d(u0_param, u0_param, (1.0 - frequency) , GMP_RNDN);
			}
			//mpfr_set_ui(v_params[item], 1, GMP_RNDN);
		}
		item_params[item] = frequency;
		//item_params[item] = 0.5;
		if (include_rowmargins) { // TODO: adjust
			//compute_all_cum_sizes();
			adjust_all_cum_sizes(item, 0.5, frequency);
			for (uint ii=0; ii<=item_cnt; ++ii)
				mpfr_set(blocks[0].node->size_array[ii], blocks[0].node->cum_size_array[ii], GMP_RNDN);
		}
	}
	
	return;
}


void ExpModel::set_rowsize_prob(int size, double prob)
{
	rowsize_probs[size] = prob;
	
	return;
}


// add an itemset to the model
// if permanent==false, it is only added temporarily to compute its frequency, not everything has to be computed 
void ExpModel::add_itemset(int *itemset, int len, double freq, bool permanent)
{
	static const bool CHECK_SIZE = true;
	
	if (CHECK_SIZE && (itemsets_count == max_k || (partition_size<<1) > max_partition_size)) {
		cout << "Size too big " << itemsets_count << ' ' <<  max_k << '\t' << (partition_size<<1) << ' ' <<  max_partition_size << endl;
		throw exception(); 
	}
	
	if (permanent) {
		//cout << '\t' << '[';
		//for (int i=0; i<len; ++i) cout << itemset[i] << ' ';
		//cout << ']' << ' ' << freq << endl;
		itemsets[itemsets_count] = new Itemset(itemset, len);
		sort(itemsets[itemsets_count]->itemset, itemsets[itemsets_count]->itemset+len); // might not be sorted, but is this uberhaupt necessary?
		frequencies[itemsets_count] = freq;
		mpfr_set_ui(u_params[itemsets_count], 1, GMP_RNDN);
	}
	else {
		dummy_itemset_adder->itemset = itemset;
		dummy_itemset_adder->len     = len;
		itemsets[itemsets_count]     = dummy_itemset_adder;
	}
	
	itemsets_bits[itemsets_count]->clear();
	bool newitems = false;
	for (int i=0; i<len; ++i) {
		itemsets_bits[itemsets_count]->set_bit(itemset[i]); // TODO: is this necessary for !permanent ?
		if (permanent) {
			if ( ! itemsets_union_bits->isset(itemset[i]) ) { // new item added; note: only happens if permanent==true
				newitems = true;
				itemsets_union_bits->set_bit(itemset[i]);
				itemsets_union[itemsets_union_size] = itemset[i];
				++itemsets_union_size;
				sort(itemsets_union, itemsets_union+itemsets_union_size);
				// TODO: init w indep rather than 0.5, see set_item_freq()
				if (include_items) {
					if (!include_rowmargins) {
						mpfr_mul_d(u0_param, u0_param, (1.0 - item_frequencies[itemset[i]]), GMP_RNDN);
						mpfr_mul_d(v0_param, v0_param, (1.0 - item_frequencies[itemset[i]]), GMP_RNDN);
						mpfr_set_d(v_params[itemset[i]], item_frequencies[itemset[i]] / (1.0 - item_frequencies[itemset[i]]), GMP_RNDN);
						
//						mpfr_mul_d(u0_param, u0_param, (1.0 - item_frequencies[itemset[i]]) / (1.0 - item_params[itemset[i]]), GMP_RNDN);
//						mpfr_mul_d(v0_param, v0_param, (1.0 - item_frequencies[itemset[i]]) / (1.0 - item_params[itemset[i]]), GMP_RNDN);
//						mpfr_set_d(v_params[itemset[i]], item_frequencies[itemset[i]] / (1.0 - item_frequencies[itemset[i]]), GMP_RNDN);
//						//mpfr_mul_d(u0_param, u0_param, item_params[itemset[i]], GMP_RNDN);
//						//mpfr_mul_d(v0_param, v0_param, 1.0-item_params[itemset[i]], GMP_RNDN);
					}
					else {
						// item already in there
//						//compute_all_cum_sizes(); // TODO: faster implementation here, maybe only once for all new items together
//						for (uint i=0; i<partition_size; ++i)
//							adjust_cum_sizes(blocks[i].node, 0.0, 0.5);
					}
				}
				else {
					if (!include_rowmargins) {
						mpfr_mul_d(u0_param, u0_param, 0.5, GMP_RNDN);
						mpfr_mul_d(v0_param, v0_param, 0.5, GMP_RNDN);
					}
				}
			}
		}
	}
	
//	if (include_rowmargins && newitems) {
//		compute_all_cum_sizes(); // TODO: adjust in stead of recompute; O(kt) rather than O(k^2), with t=nr of new items
//	}
	
	// init new blocks if we don't already have enough init'ed ones
	for (uint i=inited_blocks; i < (partition_size<<1); ++i) {
		blocks[i].node = new PartitionNode();
		++inited_blocks;
	}
	
	++itemsets_count;
	
	refine_partition(permanent);
	
	return;
}


// refine the partition when an itemset has been added
void ExpModel::refine_partition(bool permanent)
{
	uint new_partition_size = partition_size;
	uint ind0, ind1;
	
	int * new_set_b = itemsets[itemsets_count-1]->itemset;
	const int * new_set_e = new_set_b + itemsets[itemsets_count-1]->len;
	
	for (uint i=0; i<partition_size; ++i)
	{
		ind0 = blocks[i].index;
		ind1 = (ind0 << 1) | (uint)1;
		blocks[new_partition_size].index = ind1;
		PartitionNode * thisblocknode = blocks[i].node;
		PartitionNode * newblocknode  = blocks[new_partition_size].node;
		
		newblocknode->copy_fixed(*thisblocknode);
		
		// compute new cumulatives cum_size, and initialize size with them
		// if include_items==false, init size directly
		
		if (include_items) {
			
			if (include_rowmargins) {
				for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii)
					mpfr_set(newblocknode->cum_size_array[ii], thisblocknode->cum_size_array[ii], GMP_RNDN);
				//for (uint ii=itemsets_union_size+1; ii<=item_cnt; ++ii)
				//	mpfr_set_ui(newblocknode->cum_size_array[ii], 0, GMP_RNDN);
				mpfr_set(newblocknode->cum_size, thisblocknode->cum_size, GMP_RNDN);
			}
			else
				mpfr_set(newblocknode->cum_size, thisblocknode->cum_size, GMP_RNDN);
			
			for (int *iset=new_set_b; iset<new_set_e; ++iset) {
				int item = *iset;
				if (!newblocknode->isfixed(item)) {
					newblocknode->set_fixed_bit(item);
					if (include_rowmargins) {
						adjust_cum_sizes(newblocknode, item_params[item], 1.0);
						for (uint ii=0; ii<=item_cnt; ++ii)
							mpfr_mul_d(newblocknode->cum_size_array[ii], newblocknode->cum_size_array[ii], item_params[item], GMP_RNDN);
					}
					else 
						mpfr_mul_d(newblocknode->cum_size, newblocknode->cum_size, item_params[item], GMP_RNDN);
				}
			}
			
			if (include_rowmargins) {
				mpfr_set_ui(newblocknode->cum_size, 0, GMP_RNDN);
				for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii) {
					mpfr_set(newblocknode->size_array[ii], newblocknode->cum_size_array[ii], GMP_RNDN);
					mpfr_mul(tmp, w_params[ii], newblocknode->cum_size_array[ii], GMP_RNDN);
					mpfr_add(newblocknode->cum_size, newblocknode->cum_size, tmp, GMP_RNDN);
				}
			}
			mpfr_set(newblocknode->size, newblocknode->cum_size, GMP_RNDN);
			
			newblocknode->count_fixed();
		}
		else {
			if (include_rowmargins) {
				for (uint ii=0; ii<=item_cnt; ++ii)
					mpfr_set(newblocknode->cum_size_array[ii], thisblocknode->cum_size_array[ii], GMP_RNDN);
				mpfr_set(newblocknode->cum_size, thisblocknode->cum_size, GMP_RNDN);
				
				for (int *iset=new_set_b; iset<new_set_e; ++iset) {
					int item = *iset;
					if (!newblocknode->isfixed(item)) {
						newblocknode->set_fixed_bit(item);
						adjust_cum_sizes(newblocknode, 0.5, 1.0);
						for (uint ii=0; ii<=item_cnt; ++ii)
							mpfr_mul_d(newblocknode->cum_size_array[ii], newblocknode->cum_size_array[ii], 0.5, GMP_RNDN);
					}
				}
				
				mpfr_set_ui(newblocknode->cum_size, 0, GMP_RNDN);
				for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii) {
					mpfr_set(newblocknode->size_array[ii], newblocknode->cum_size_array[ii], GMP_RNDN);
					mpfr_mul(tmp, w_params[ii], newblocknode->cum_size_array[ii], GMP_RNDN);
					mpfr_add(newblocknode->cum_size, newblocknode->cum_size, tmp, GMP_RNDN);
				}
				mpfr_set(newblocknode->size, newblocknode->cum_size, GMP_RNDN);
				
				newblocknode->count_fixed();
			}
			else {
				for (int *iset=new_set_b; iset<new_set_e; ++iset) {
					newblocknode->set_fixed_bit(*iset);
				}
				newblocknode->count_fixed();
				
				mpfr_set_ui(newblocknode->size, 1, GMP_RNDN);
				mpfr_mul_2si(newblocknode->size, newblocknode->size, -newblocknode->fixed.bit_cnt, GMP_RNDN);
			}
		}
		
		++new_partition_size;
	}
	
	if (permanent) {
		for (uint i=0; i<partition_size; ++i)
			blocks[i].index <<= 1;
		if (!include_rowmargins) {
			compute_block_sizes();
			for (uint i=0; i<partition_size; ++i)
				mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocks[i+partition_size].node->size, GMP_RNDN);
		}
		else {
			compute_block_size_arrays();
			for (uint i=0; i<partition_size; ++i) {
				mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocks[i+partition_size].node->size, GMP_RNDN);
				for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii)
					mpfr_sub(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], blocks[i+partition_size].node->size_array[ii], GMP_RNDN);
			}
			
//			// compute the size field, which should contain the full probability of the current block with respect to the background info
//			for (uint i=0; i<partition_size*2; ++i) {
//				mpfr_set_ui(blocks[i].node->size, 0, GMP_RNDN);
//				for (uint ii=0; ii<=item_cnt; ++ii) {
//					mpfr_mul(tmp, w_params[ii], blocks[i].node->size_array[ii], GMP_RNDN);
//					mpfr_add(blocks[i].node->size, blocks[i].node->size, tmp, GMP_RNDN);
//				}
//			}
		}
		compute_block_probs(permanent);
		partition_size = new_partition_size;
		interleave_blocks(0, partition_size);
		//prune_blocks_if_empty();
	}
	else {
		if (include_rowmargins) {
			compute_all_cum_sizes(); // TODO: seems a bit much, ey?
			compute_block_size_arrays();
//			// compute the size field, which should contain the full probability of the current block with respect to the background info
//			for (uint i=partition_size; i<partition_size*2; ++i) {
//				mpfr_set_ui(blocks[i].node->size, 0, GMP_RNDN);
//				for (uint ii=0; ii<=item_cnt; ++ii) {
//					mpfr_mul(tmp, w_params[ii], blocks[i].node->size_array[ii], GMP_RNDN);
//					mpfr_add(blocks[i].node->size, blocks[i].node->size, tmp, GMP_RNDN);
//				}
//			}
		}
		else
			compute_block_sizes();
		compute_block_probs(permanent);
		partition_size = new_partition_size;
	}
	
	return;
}


// note: size is ~ probability according to uniform or indep distribution
// compute sizes from cululative sizes
void ExpModel::compute_block_sizes()
{
	uint mask;
	//uint from = partition_size;
	uint to = partition_size << 1;
	if (partition_size == ((uint)1<<(itemsets_count-1))) { // we can do plain old QIE
		PartitionBlock * blocksplus = blocks + partition_size;
		for (uint j=0; j<itemsets_count-1; ++j) {
			mask = (uint)1 << j;
			for (uint i=partition_size; i<to; ++i) {
				uint ind0 = blocks[i].index >> 1;
				if ( (ind0 & mask) == 0 ) {
					mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocksplus[ind0 | mask].node->size, GMP_RNDN);
				}
			}
		}
	}
//	else if (!DO_QIE) { // quadratic in partition_size
//		for (uint i=to-1; i>=partition_size; --i) {
//			uint ind = blocks[i].index;
//			for (uint j=i+1; j<to; ++j) {
//				if ((ind & blocks[j].index) == ind )
//					mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocks[j].node->size, GMP_RNDN);
//			}
//		}
//	}
	else { // genaralized QIE, O(itemsets_count * partition_size)
		if (itemsets_count > 1)
			qie(0, partition_size, to);
	}
	
	// set the sizes where k-th bit == 0, no need to look them up, we know they are always present
	// also, sizes in the lower half are not powers of two
	// note: this is moved out
//	for (uint i=0; i<partition_size; ++i)
//		mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocks[i+partition_size].node->size, GMP_RNDN);
	
	return;
}


void ExpModel::compute_block_size_arrays()
{
	// TODO: make partial arrangements when adding permanent itemset
	
	uint mask;
	uint to = partition_size << 1;
	if (partition_size == ((uint)1<<(itemsets_count-1))) { // we can do plain old QIE
		PartitionBlock * blocksplus = blocks + partition_size;
		for (uint j=0; j<itemsets_count-1; ++j) {
			mask = (uint)1 << j;
			for (uint i=partition_size; i<to; ++i) {
				uint ind0 = blocks[i].index >> 1;
				if ( (ind0 & mask) == 0 ) {
					mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocksplus[ind0 | mask].node->size, GMP_RNDN);
					for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii)
						mpfr_sub(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], blocksplus[ind0 | mask].node->size_array[ii], GMP_RNDN);
				}
			}
		}
	}
	else { // genaralized QIE, O(itemsets_count * partition_size)
		if (itemsets_count > 1)
			qie_size_array(0, partition_size, to);
	}
	
	// set the sizes where k-th bit == 0, no need to look them up, we know they are always present
	// also, sizes in the lower half are not powers of two
	// note: moved out
	//for (uint i=0; i<partition_size; ++i)
	//	for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii)
	//		mpfr_sub(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], blocks[i+partition_size].node->size_array[ii], GMP_RNDN);
	
	return;
}



void ExpModel::qie(uint j, uint lo, uint hi)
{
	if (lo==hi)
		return;
	
	const uint mask = (uint)1 << (itemsets_count-j-1);
	uint mid = get_first_mask_index(lo, hi, mask);
	
	uint loc = mid;
	
	if (mid < hi) {
		for (uint i=lo; i<mid; ++i) {
			uint ind0 = blocks[i].index;
			uint ind1 = ind0 | mask;
			if (blocks[loc].index != ind1) {
				loc = get_index(mid, hi, ind1); // TODO: loc = mid+i-lo; -ish ?
				if (loc == hi) {
					ind1 = prefix_closure(ind1, j, blocks[i].node);
					if (ind1 != (ind0|mask))
						loc = get_index(mid, hi, ind1);
				}
			}
			if (loc != hi) {
				mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocks[loc].node->size, GMP_RNDN);
				++loc;
			}
		}
	}
	
	if (j < itemsets_count-2) {
		qie(j+1, lo, mid);
		qie(j+1, mid, hi);
	}
	
	return;
}


uint ExpModel::prefix_closure(const uint vect, const uint j, const PartitionNode *bloxnode)
{
	uint closure = vect;
	bitvect->copy_bits(bloxnode->fixed);
	bitvect->update(*(itemsets_bits[j]));
	
	// TODO: invert jj and maxbit into j and minbit
	const uint jj = itemsets_count-j-1;
	uint maxbit = 0;
	uint mask = 1;
	while ((maxbit<jj) && ((closure & mask)==0)) { ++maxbit; mask<<=1; }
	
	if (maxbit==jj) return closure;
	
	for (uint i=itemsets_count-maxbit-2; i>j; --i) {
		mask <<= 1;
		if ( (closure & mask) == 0) {
			if (bitvect->contains(*(itemsets_bits[i])))
				closure |= mask;
		}
	}
	
	return closure;
}


void ExpModel::qie_size_array(uint j, uint lo, uint hi)
{
	if (lo==hi)
		return;
	
	const uint mask = (uint)1 << (itemsets_count-j-1);
	uint mid = get_first_mask_index(lo, hi, mask);
	
	uint loc = mid;
	
	if (mid < hi) {
		for (uint i=lo; i<mid; ++i) {
			uint ind0 = blocks[i].index;
			uint ind1 = ind0 | mask;
			if (blocks[loc].index != ind1) {
				loc = get_index(mid, hi, ind1); // TODO: loc = mid+i-lo; -ish ?
				if (loc == hi) {
					ind1 = prefix_closure(ind1, j, blocks[i].node);
					if (ind1 != (ind0|mask))
						loc = get_index(mid, hi, ind1);
				}
			}
			if (loc != hi) {
				mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocks[loc].node->size, GMP_RNDN);
				for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii)
					mpfr_sub(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], blocks[loc].node->size_array[ii], GMP_RNDN);
				++loc;
			}
		}
	}
	
	if (j < itemsets_count-2) {
		qie_size_array(j+1, lo, mid);
		qie_size_array(j+1, mid, hi);
	}
	
	return;
}


// In-place linear time interleaving, aka 'in-shuffle', e.g., abc123 -> a1b2c3. Really cool.
void ExpModel::interleave_blocks(uint offset, uint len)
{
	//if (len==0)
	//	return;
	
	uint n = len / 2;
	uint k = uint(floor(log(len+1)/log(3)));
	uint m2 = uint((pow(3, k) - 1));
	
	reverse_blockrange(offset + m2/2, offset + m2/2+n);
	reverse_blockrange(offset + m2/2, offset + m2);
	reverse_blockrange(offset + m2, offset + n+m2/2);
	
	// cycle leader algorithm
	uint leader = 1;
	for (uint i=0; i<k; ++i) { 
		uint j = leader;
		do {
			j = (2 * j) % (m2 + 1);
			swap(blocks[leader+offset-1].index, blocks[j+offset-1].index);
			swap(blocks[leader+offset-1].node,  blocks[j+offset-1].node);
		} while (j != leader);
		leader *= 3;
	}
	
	//change interleaving 1a2b3c -> a1b2c3
	for (uint i=0; i<m2; i+=2) {
		swap(blocks[offset+i].index, blocks[offset+i+1].index);
		swap(blocks[offset+i].node,  blocks[offset+i+1].node);
	}
	
	if (len > m2)
		interleave_blocks(offset+m2, len-m2);
	
	return;
}


void ExpModel::compute_block_probs(bool permanent)
{
	if (permanent) {
		for (uint i=0; i<partition_size; ++i) {
			PartitionNode * ni = blocks[i].node;
			PartitionNode * nj = blocks[i + partition_size].node;
			if (mpfr_sgn(ni->prob) <= 0) {
				mpfr_set_ui(nj->prob, 0, GMP_RNDN);
				continue;
			}
			mpfr_add(nj->prob, ni->size, nj->size, GMP_RNDN);
			mpfr_div(nj->prob, nj->size, nj->prob, GMP_RNDN);
			mpfr_mul(nj->prob, ni->prob, nj->prob, GMP_RNDN);
			mpfr_sub(ni->prob, ni->prob, nj->prob, GMP_RNDN);
		}
	}
	else {
		for (uint i=0; i<partition_size; ++i) {
			PartitionNode * ni = blocks[i].node;
			PartitionNode * nj = blocks[i + partition_size].node;
			if (mpfr_sgn(ni->prob) <= 0) {
				mpfr_set_ui(nj->prob, 0, GMP_RNDN);
				continue;
			}
			mpfr_div(nj->prob, nj->size, ni->size, GMP_RNDN);
			mpfr_mul(nj->prob, ni->prob, nj->prob, GMP_RNDN);
		}
	}
	
	return;
}


// TODO: skip 0 assignment to blocks that are already zero, perhaps add a bool?
void ExpModel::prune_blocks_if_empty()
{
	uint gap_count = 0;
	for (uint i=0; i<partition_size; ++i) {
		const bool may_prune_if_empty = !DO_QIE || (blocks[i].index & (uint)1); // if index ends with 0, do not prune
		if (mpfr_sgn(blocks[i].node->size) <= 0) { 
			if (may_prune_if_empty) {
				++gap_count;
			}
			else {
				if (gap_count) {
					blocks[i - gap_count].index = blocks[i].index;
					swap(blocks[i - gap_count].node, blocks[i].node);
				}
				mpfr_set_ui(blocks[i-gap_count].node->size, 0, GMP_RNDN);
				mpfr_set_ui(blocks[i-gap_count].node->prob, 0, GMP_RNDN);
			}
		}
		else {
			if (gap_count) {
				blocks[i - gap_count].index = blocks[i].index;
				swap(blocks[i - gap_count].node, blocks[i].node);
			}
		}
	}
	
	partition_size -= gap_count;
	
	return;
}


void ExpModel::prune_blocks_if_prob_zero()
{
	uint gap_count = 0;
	for (uint i=0; i<partition_size; ++i) {
		const bool may_prune_if_empty = !DO_QIE || (blocks[i].index & (uint)1); // if index ends with 0, do not prune
		if (mpfr_sgn(blocks[i].node->prob) <= 0) {
			if (may_prune_if_empty) {
				++gap_count;
			}
			else {
				if (gap_count) {
					blocks[i - gap_count].index = blocks[i].index;
					swap(blocks[i - gap_count].node, blocks[i].node);
				}
				mpfr_set_ui(blocks[i-gap_count].node->prob, 0, GMP_RNDN);
			}
		}
		else {
			if (gap_count) {
				blocks[i - gap_count].index = blocks[i].index;
				swap(blocks[i - gap_count].node, blocks[i].node);
			}
		}
	}
	
	partition_size -= gap_count;
	
	return;
}


void ExpModel::adjust_all_cum_sizes(int item, double oldparam, double newparam)
{
	if (include_rowmargins) {
		for (uint i=0; i<partition_size; ++i) {
			if (blocks[i].node->isfixed(item)) {
				for (uint ii=0; ii<=item_cnt; ++ii)
					mpfr_mul_d(blocks[i].node->cum_size_array[ii], blocks[i].node->cum_size_array[ii], newparam/oldparam, GMP_RNDN);
				mpfr_mul_d(blocks[i].node->cum_size, blocks[i].node->cum_size, newparam/oldparam, GMP_RNDN);
			}
			else {
				adjust_cum_sizes(blocks[i].node, oldparam, newparam);
				mpfr_set_ui(blocks[i].node->cum_size, 0, GMP_RNDN);
				for (uint ii=0; ii<=item_cnt; ++ii) {
					mpfr_mul(tmp, w_params[ii], blocks[i].node->cum_size_array[ii], GMP_RNDN);
					mpfr_add(blocks[i].node->cum_size, blocks[i].node->cum_size, tmp, GMP_RNDN);
				}
			}
		}
	}
	else {
		for (uint i=0; i<partition_size; ++i) {
			if (blocks[i].node->isfixed(item))
				mpfr_mul_d(blocks[i].node->cum_size, blocks[i].node->cum_size, newparam/oldparam, GMP_RNDN);
		}
	}
	
	return;
}


void ExpModel::adjust_cum_sizes(PartitionNode * blocknode, double oldparam, double newparam)
{
//	mpfr_set_ui(factor, 1, GMP_RNDN);
//	for (uint i=0; i<item_cnt; ++i)
//		if (blocknode->isfixed(i))
//			mpfr_mul_d(factor, factor, item_params[i]);
//	for (uint ii=0; ii<=item_cnt; ++ii)
//		mpfr_div(blocknode->cum_size_array[i], blocknode->cum_size_array[i], factor, GMP_RNDN);
	
	// remove item
	mpfr_div_d(blocknode->cum_size_array[0], blocknode->cum_size_array[0], 1.0-oldparam, GMP_RNDN);
	for (uint ii=1; ii<=/*itemsets_union_size*/item_cnt; ++ii) {
		mpfr_mul_d(tmp, blocknode->cum_size_array[ii-1], oldparam, GMP_RNDN);
		mpfr_sub(blocknode->cum_size_array[ii], blocknode->cum_size_array[ii], tmp, GMP_RNDN);
		mpfr_div_d(blocknode->cum_size_array[ii], blocknode->cum_size_array[ii], 1.0-oldparam, GMP_RNDN);
	}
	
	// re-add item
	if (newparam == 1.0) {
		for (uint ii=/*itemsets_union_size*/item_cnt; ii>0; --ii)
			mpfr_set(blocknode->cum_size_array[ii], blocknode->cum_size_array[ii-1], GMP_RNDN);
		mpfr_set_ui(blocknode->cum_size_array[0], 0, GMP_RNDN);
	}
	else {
		//if (itemsets_union_size>0) {
			//mpfr_mul_d(blocknode->cum_size_array[itemsets_union_size], blocknode->cum_size_array[itemsets_union_size-1], newparam, GMP_RNDN);
			mpfr_mul_d(blocknode->cum_size_array[item_cnt], blocknode->cum_size_array[item_cnt-1], newparam, GMP_RNDN);
			for (uint ii=/*itemsets_union_size*/item_cnt-1; ii>0; --ii) {
				mpfr_mul_d(tmp, blocknode->cum_size_array[ii-1], newparam, GMP_RNDN);
				mpfr_mul_d(blocknode->cum_size_array[ii], blocknode->cum_size_array[ii], 1.0-newparam, GMP_RNDN);
				mpfr_add(blocknode->cum_size_array[ii], blocknode->cum_size_array[ii], tmp, GMP_RNDN);
			}
		//}
		mpfr_mul_d(blocknode->cum_size_array[0], blocknode->cum_size_array[0], 1.0-newparam, GMP_RNDN);
	}
	
//	for (uint ii=0; ii<=item_cnt; ++ii)
//		mpfr_mul(blocknode->cum_size_array[i], blocknode->cum_size_array[i], factor, GMP_RNDN);
	
	// sum it up
	mpfr_set_ui(blocknode->cum_size, 0, GMP_RNDN);
	for (uint ii=0; ii<=itemsets_union_size; ++ii) {
		mpfr_mul(tmp, blocknode->cum_size_array[ii], w_params[ii], GMP_RNDN);
		mpfr_add(blocknode->cum_size, blocknode->cum_size, tmp, GMP_RNDN);
	}
	
	return;
}


void ExpModel::compute_all_cum_sizes()
{
	if (include_rowmargins) {
		if (include_items) {
			// this is quadratic because from scratch
			for (uint i=0; i<partition_size; ++i) {
				
				PartitionNode * bn = blocks[i].node;
				mpfr_t *csa = bn->cum_size_array;
				mpfr_set_ui(csa[0], 1, GMP_RNDN);
				
				for (uint j=0; j</*itemsets_union_size*/item_cnt; ++j) {
					//int jj = itemsets_union[j];
					int jj = j;
					if (bn->isfixed(jj)) {
						for (uint ii=j+1; ii>0; --ii)
							mpfr_set(csa[ii], csa[ii-1], GMP_RNDN);
						mpfr_set_ui(csa[0], 0, GMP_RNDN);
					}
					else {
						mpfr_mul_d(csa[j+1], csa[j], item_params[jj], GMP_RNDN);
						for (uint ii=j; ii>0; --ii) {
							mpfr_mul_d(factor, csa[ii-1], item_params[jj], GMP_RNDN);
							mpfr_mul_d(csa[ii], csa[ii], 1.0-item_params[jj], GMP_RNDN);
							mpfr_add(csa[ii], csa[ii], factor, GMP_RNDN);
						}
						mpfr_mul_d(csa[0], csa[0], 1.0-item_params[jj], GMP_RNDN);
					}
				}
				
				mpfr_set_ui(factor, 1, GMP_RNDN);
				for (uint j=0; j<itemsets_union_size; ++j) {
					if (bn->isfixed(itemsets_union[j]))
						mpfr_mul_d(factor, factor, item_params[itemsets_union[j]], GMP_RNDN);
				}
				for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii)
					mpfr_mul(csa[ii], csa[ii], factor, GMP_RNDN);
				
				mpfr_set_ui(bn->cum_size, 0, GMP_RNDN);
				for (uint ii=0; ii<=/*itemsets_union_size*/item_cnt; ++ii) {
					mpfr_mul(tmp, w_params[ii], csa[ii], GMP_RNDN);
					mpfr_add(bn->cum_size, bn->cum_size, tmp, GMP_RNDN);
				}
				
			}
		}
		else {
			for (uint i=0; i<partition_size; ++i) {
				mpfr_t *csa = blocks[i].node->cum_size_array;
				uint fixed = blocks[i].node->fixed.bit_cnt;
				for (uint ii=0; ii<fixed; ++ii)
					mpfr_set_ui(csa[ii], 0, GMP_RNDN);
				mpfr_set_ui(csa[fixed], 1, GMP_RNDN);
				for (uint ii=fixed+1; ii<=/*itemsets_union_size*/item_cnt; ++ii) {
					mpfr_mul_ui(csa[ii], csa[ii-1], /*itemsets_union_size*/item_cnt-ii+1, GMP_RNDN);
					mpfr_div_ui(csa[ii], csa[ii], ii-fixed, GMP_RNDN);
				}
				for (uint ii=fixed; ii<=/*itemsets_union_size*/item_cnt; ++ii)
					mpfr_mul_2si(csa[ii], csa[ii], (int)fixed-/*itemsets_union_size*/(int)item_cnt, GMP_RNDN);
				mpfr_set_ui(blocks[i].node->cum_size, 0, GMP_RNDN);
				for (uint ii=fixed; ii<=/*itemsets_union_size*/item_cnt; ++ii) {
					mpfr_mul(tmp, w_params[ii], csa[ii], GMP_RNDN);
					mpfr_add(blocks[i].node->cum_size, blocks[i].node->cum_size, tmp, GMP_RNDN);
				}
			}

		}
	}
	else {
		for (uint i=0; i<partition_size; ++i) {
			mpfr_set_ui(blocks[i].node->cum_size, 1, GMP_RNDN);
			for (uint j=0; j<itemsets_union_size; ++j) {
				if (blocks[i].node->isfixed(itemsets_union[j]))
					mpfr_mul_d(blocks[i].node->cum_size, blocks[i].node->cum_size, item_params[itemsets_union[j]], GMP_RNDN);
			}
		}
	}
	
	return;
}


// @pre: last itemset must have been added with permanent=false
void ExpModel::remove_last_itemset()
{
	partition_size >>= 1;
	--itemsets_count;
	
	return;
}


void ExpModel::delete_itemset(int j)
{
	// remove the itemset from itemsets and shift, same for frequencies, itemsets_bits
	// shift u parameters, adjust u0
	// adjust itemsets_union
	
	// go over blocks, collapse, redo indices
	// IMPORTANT: is TC+ property still satisfied
	
	return;
}


// @pre: itemset may only contain items that are already in the model (itemsets_union)
double ExpModel::estimate_frequency(int *itemset, int len)
{
	add_itemset(itemset, len, 0.0, false);
	
	mpfr_set_ui(estimate, 0, GMP_RNDN);
	for (uint i=(partition_size>>1); i<partition_size; ++i)
		mpfr_add(estimate, estimate, blocks[i].node->prob, GMP_RNDN);
	
	remove_last_itemset();
	
	return mpfr_get_d(estimate, GMP_RNDN);

}


void ExpModel::normalize()
{
	static const double epsilon = 1.0e-10;
	
	mpfr_set_d(sum, 0.0, GMP_RNDN);
	
	for (uint i=0; i<partition_size; ++i)
		mpfr_add(sum, sum, blocks[i].node->prob, GMP_RNDN);
	
	if (fabs(mpfr_get_d(sum, GMP_RNDN)- 1.0) > epsilon) {
		cout << "NORMALIZE " << mpfr_get_d(sum, GMP_RNDN) << endl; // if all goes well we should never see this
		for (uint i=0; i<partition_size; ++i)
			mpfr_div(blocks[i].node->prob, blocks[i].node->prob, sum, GMP_RNDN);
		
		// adjust u0?
		mpfr_div(u0_param, u0_param, sum, GMP_RNDN);
	}
	
	return;
}



double ExpModel::adjust_item_param(int item)
{
	if (item_frequencies[item] == 1.0)
		return 0.0;
	
	add_itemset(&item, 1, 0.0, false);
	
	uint middle = partition_size >> 1;
	
	mpfr_set_ui(estimate, 0, GMP_RNDN);
	for (uint i=middle; i<partition_size; ++i) {
		mpfr_add(estimate, estimate, blocks[i].node->prob, GMP_RNDN);
	}
	
	double abs_err = fabs(mpfr_get_d(estimate, GMP_RNDN)-item_frequencies[item]);
	if (abs_err < 1e-5) {
		remove_last_itemset();
		return abs_err;
	}
	else {
		for (uint i=0; i<middle; ++i) {
			mpfr_sub(blocks[i].node->size, blocks[i].node->size, blocks[i+middle].node->size, GMP_RNDN);
			mpfr_sub(blocks[i].node->prob, blocks[i].node->prob, blocks[i+middle].node->prob, GMP_RNDN);
			if (include_rowmargins)
				for (uint ii=0; ii<=item_cnt; ++ii)
					mpfr_sub(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], blocks[i+middle].node->size_array[ii], GMP_RNDN);
		}
	}
	
	// f1 = fr/est; f2=(1-fr)/(1-est)
	mpfr_d_div(f1, item_frequencies[item], estimate, GMP_RNDN);
	mpfr_ui_sub(estimate, 1, estimate, GMP_RNDN);
	mpfr_d_div(f2, 1.0 - item_frequencies[item], estimate, GMP_RNDN);
	mpfr_ui_sub(estimate, 1, estimate, GMP_RNDN); // unnecessary?
	
	// update v, u0 parameters
	mpfr_mul(u0_param, u0_param, f2, GMP_RNDN);
	mpfr_mul(v_params[item], v_params[item], f1, GMP_RNDN);
	mpfr_div(v_params[item], v_params[item], f2, GMP_RNDN);
	
	double param = item_params[item];
	
	//adjust blocks according to frequency
	
	param = 1.0 - param;
	for (uint i=0; i<middle; ++i) {
		mpfr_div_d(blocks[i].node->size, blocks[i].node->size, param, GMP_RNDN);
		if (include_rowmargins)
			for (uint ii=0; ii<=item_cnt; ++ii)
				mpfr_div_d(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], param, GMP_RNDN);
	}
	param = 1.0 - param;
	for (uint i=middle; i<partition_size; ++i) {
		mpfr_div_d(blocks[i].node->size, blocks[i].node->size, param, GMP_RNDN);
		if (include_rowmargins)
			for (uint ii=0; ii<=item_cnt; ++ii)
				mpfr_div_d(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], param, GMP_RNDN);
	}
	
	mpfr_add_ui(tmp, v_params[item], 1, GMP_RNDN);
	mpfr_div(tmp, v_params[item], tmp, GMP_RNDN);
	double newparam = mpfr_get_d(tmp, GMP_RNDN);
	item_params[item] = newparam;
	
	newparam = 1.0 - newparam;
	for (uint i=0; i<middle; ++i) {
		mpfr_mul(blocks[i].node->prob, blocks[i].node->prob, f2, GMP_RNDN);
		mpfr_mul_d(blocks[i].node->size, blocks[i].node->size, newparam, GMP_RNDN);
		if (include_rowmargins)
			for (uint ii=0; ii<=item_cnt; ++ii)
				mpfr_mul_d(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], newparam, GMP_RNDN);
	}
	newparam = 1.0 - newparam;
	for (uint i=middle; i<partition_size; ++i) {
		mpfr_mul(blocks[i].node->prob, blocks[i].node->prob, f1, GMP_RNDN);
		mpfr_mul_d(blocks[i].node->size, blocks[i].node->size, newparam, GMP_RNDN);
		if (include_rowmargins)
			for (uint ii=0; ii<=item_cnt; ++ii)
				mpfr_mul_d(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], newparam, GMP_RNDN);
	}
	
	//adjust v0
	mpfr_div_d(v0_param, v0_param, 1.0 - param, GMP_RNDN);
	mpfr_mul_d(v0_param, v0_param, 1.0 - newparam, GMP_RNDN);
	
	//marginalize back
	for (uint i=0; i<middle; ++i) {
		mpfr_add(blocks[i].node->size, blocks[i].node->size, blocks[i+middle].node->size, GMP_RNDN);
		mpfr_add(blocks[i].node->prob, blocks[i].node->prob, blocks[i+middle].node->prob, GMP_RNDN);
		if (include_rowmargins)
			for (uint ii=0; ii<=item_cnt; ++ii)
				mpfr_add(blocks[i].node->size_array[ii], blocks[i].node->size_array[ii], blocks[i+middle].node->size_array[ii], GMP_RNDN);
	}
	
	remove_last_itemset();
	
	//TODO: reinstate
	adjust_all_cum_sizes(item, param, newparam); 
	//compute_all_cum_sizes();
	
	return abs_err;
}


void ExpModel::iterative_scaling(int maxloops)
{
	static const double epsilon = 1e-5;
	double max_error = 10 * epsilon + 1;
	
	sync_params();
	
	while (maxloops-- > 0 && max_error > 10*epsilon)
	{
		max_error = 0.0;
		
		// do the itemset constraints
		for (uint j=0; j<itemsets_count; ++j)
		{
			uint ind = (uint)1 << (itemsets_count-j-1);
			
			mpfr_set_d(estimate, 0.0, GMP_RNDN);
			for (uint i=0; i<partition_size; ++i) {
				if (blocks[i].index & ind)
					mpfr_add(estimate, estimate, blocks[i].node->prob, GMP_RNDN);
			}
			
			max_error = max(max_error, fabs(mpfr_get_d(estimate, GMP_RNDN) - frequencies[j]));
			
			mpfr_d_div(f1, frequencies[j], estimate, GMP_RNDN);
			mpfr_ui_sub(estimate, 1, estimate, GMP_RNDN);
			mpfr_d_div(f2, 1.0-frequencies[j], estimate, GMP_RNDN);
			//mpfr_ui_sub(estimate, 1, estimate, GMP_RNDN);
			
			mpfr_mul(u_params[j], u_params[j], f1, GMP_RNDN);
			mpfr_div(u_params[j], u_params[j], f2, GMP_RNDN);
			mpfr_mul(u0_param, u0_param, f2, GMP_RNDN);
			
			for (uint i=0; i<partition_size; ++i) {
				if (blocks[i].index & ind)
					mpfr_mul(blocks[i].node->prob, blocks[i].node->prob, f1, GMP_RNDN);
				else
					mpfr_mul(blocks[i].node->prob, blocks[i].node->prob, f2, GMP_RNDN);
			}
			
		}
		
		// do the item constraints
		if (include_items) {
			if (include_rowmargins) {
				for (uint j=0; j<item_cnt; ++j) {
					max_error = max(max_error, adjust_item_param(j));
				}
			} else {
				for (uint j=0; j<itemsets_union_size; ++j) {
					max_error = max(max_error, adjust_item_param(itemsets_union[j]));
				}
			}
			normalize();
			//compute_all_cum_sizes();
		}
		
		// do the rowsize constraints
		if (include_rowmargins) {
			for (uint ii=0; ii<=item_cnt; ++ii)
				mpfr_set_ui(rowsize_estimates[ii], 0, GMP_RNDN);
			
			for (uint i=0; i<partition_size; ++i) {
				if (mpfr_sgn(blocks[i].node->size) > 0) {
					mpfr_div(factor, blocks[i].node->prob, blocks[i].node->size, GMP_RNDN);
					for (uint ii=0; ii<=item_cnt; ++ii) {
						mpfr_mul(tmp, factor, blocks[i].node->size_array[ii], GMP_RNDN);
						mpfr_mul(tmp, tmp, w_params[ii], GMP_RNDN);
						mpfr_add(rowsize_estimates[ii], rowsize_estimates[ii], tmp, GMP_RNDN);
					}
				}
			}
			
			for (uint ii=0; ii<=item_cnt; ++ii) {
				double est_p = mpfr_get_d(rowsize_estimates[ii], GMP_RNDN);
				max_error = max(max_error, fabs(est_p-rowsize_probs[ii]));
				if (rowsize_probs[ii]==0.0)
					mpfr_set_ui(w_params[ii], 0, GMP_RNDN);
				else
					mpfr_mul_d(w_params[ii], w_params[ii], rowsize_probs[ii]/est_p, GMP_RNDN);
			}
			
			for (uint i=0; i<partition_size; ++i) {
				if (mpfr_sgn(blocks[i].node->size) > 0) {
					mpfr_div(blocks[i].node->prob, blocks[i].node->prob, blocks[i].node->size, GMP_RNDN);
					mpfr_set_ui(blocks[i].node->size, 0, GMP_RNDN);
					mpfr_set_ui(blocks[i].node->cum_size, 0, GMP_RNDN);
					for (uint ii=0; ii<=item_cnt; ++ii) {
						mpfr_mul(tmp, w_params[ii], blocks[i].node->size_array[ii], GMP_RNDN);
						mpfr_add(blocks[i].node->size, blocks[i].node->size, tmp, GMP_RNDN);
						mpfr_mul(tmp, w_params[ii], blocks[i].node->cum_size_array[ii], GMP_RNDN);
						mpfr_add(blocks[i].node->cum_size, blocks[i].node->cum_size, tmp, GMP_RNDN);
					}
					mpfr_mul(blocks[i].node->prob, blocks[i].node->prob, blocks[i].node->size, GMP_RNDN);
				}
			}
			//compute_all_cum_sizes();
			
		}
		
		if (max_error < epsilon)
			break;
	}
	
	sync_params(); // includes a call normalize()
	
	return;
}


void ExpModel::sync_params()
{
	if (include_items) {
		if (include_rowmargins) {
			// v_params to item_params, adjust v0_param
			mpfr_set_ui(v0_param, 1, GMP_RNDN);
			for (uint i=0; i<item_cnt; ++i) {
				item_params[i] = mpfr_get_d(v_params[i], GMP_RNDN) / (1.0 + mpfr_get_d(v_params[i], GMP_RNDN));
				//mpfr_add_ui(tmp, v_params[i], 1, GMP_RNDN);
				//mpfr_div(v0_param, v0_param, tmp, GMP_RNDN);
				mpfr_mul_d(v0_param, v0_param, 1.0-item_params[i], GMP_RNDN);
			}
		}
		else {
			// v_params to item_params, adjust v0_param
			// see note in adjust_item_param()
			mpfr_set_ui(v0_param, 1, GMP_RNDN);
			//mpfr_mul_2si(v0_param, v0_param, (int)itemsets_union_size-(int)item_cnt, GMP_RNDN);
			for (uint i=0; i<itemsets_union_size; ++i) {
				int item = itemsets_union[i];
				item_params[item] = mpfr_get_d(v_params[item], GMP_RNDN) / (1.0 + mpfr_get_d(v_params[item], GMP_RNDN));
				//mpfr_add_ui(tmp, v_params[item], 1, GMP_RNDN);
				//mpfr_div(v0_param, v0_param, tmp, GMP_RNDN);
				mpfr_mul_d(v0_param, v0_param, 1.0-item_params[item], GMP_RNDN);
			}
		}
	}
	
	// u_params to .probs
	mpfr_div(factor, u0_param, v0_param, GMP_RNDN);
	for (uint i=0; i<partition_size; ++i) {
		mpfr_set(f, factor, GMP_RNDN);
		for (uint j=0; j<itemsets_count; ++j) {
			if (blocks[i].index & ((uint)1 << (itemsets_count-j-1)))
				mpfr_mul(f, f, u_params[j], GMP_RNDN);
		}
		mpfr_mul(blocks[i].node->prob, blocks[i].node->size, f, GMP_RNDN);
	}
	
	normalize();
	
	return;
}


double ExpModel::get_entropy()
{
	mpfr_set_d(h, 0.0, GMP_RNDN);
	
	mpfr_log2(f, u0_param, GMP_RNDN);
	mpfr_sub(h, h, f, GMP_RNDN);
	
	for (uint i=0; i<itemsets_count; ++i) {
		mpfr_log2(f, u_params[i], GMP_RNDN);
		mpfr_mul_d(f, f, frequencies[i], GMP_RNDN);
		mpfr_sub(h, h, f, GMP_RNDN);
	}
	
	if (include_rowmargins) {
		if (include_items) {
			for (uint j=0; j<item_cnt; ++j) {
				mpfr_log2(f, v_params[j], GMP_RNDN);
				mpfr_mul_d(f, f, item_frequencies[j], GMP_RNDN);
				mpfr_sub(h, h, f, GMP_RNDN);
			}
		}
		for (uint ii=0; ii<=item_cnt; ++ii) {
			if (mpfr_sgn(w_params[ii]) > 0) {
				mpfr_log2(f, w_params[ii], GMP_RNDN);
				mpfr_mul_d(f, f, rowsize_probs[ii], GMP_RNDN);
				mpfr_sub(h, h, f, GMP_RNDN);
			}
		}
	}
	else {
		if (include_items) {
			for (uint i=0; i<itemsets_union_size; ++i) {
				mpfr_log2(f, v_params[itemsets_union[i]], GMP_RNDN);
				mpfr_mul_d(f, f, item_frequencies[itemsets_union[i]], GMP_RNDN);
				mpfr_sub(h, h, f, GMP_RNDN);
			}
		}
	}
	
	if (mpfr_nan_p(h)) {
		cout << mpfr_get_d(u0_param, GMP_RNDN) << "  ";
		for (uint i=0; i<itemsets_count; ++i)
			cout << mpfr_get_d(u_params[i], GMP_RNDN) << "  ";
		cout << endl;
		if (include_rowmargins) {
			if (include_items)
				for (uint j=0; j<item_cnt; ++j)
					cout << mpfr_get_d(v_params[j], GMP_RNDN) << "  ";
			cout << endl;
			for (uint ii=0; ii<=item_cnt; ++ii)
				cout << mpfr_get_d(w_params[ii], GMP_RNDN) << "  ";
			cout << endl;
		}
		else {
			if (include_items)
				for (uint j=0; j<itemsets_union_size; ++j)
					cout << mpfr_get_d(v_params[itemsets_union[j]], GMP_RNDN) << "  ";
			cout << endl;
		}
	}
	
	return mpfr_get_d(h, GMP_RNDN);
}



void ExpModel::print_model()
{
	printf("--------\n");
	printf("SETS: %i\tBLOCKS: %i\n", itemsets_count, partition_size);
	for (uint i=0; i<partition_size; ++i) {
		printf("%3u ", blocks[i].index);
		for (uint j=0; j<itemsets_count; ++j)
			printf("%u", (blocks[i].index >> (itemsets_count-j-1)) % 2);
		printf("\t%.8f\t%.8f", mpfr_get_d(blocks[i].node->size, GMP_RNDN), mpfr_get_d(blocks[i].node->prob, GMP_RNDN));
		if (include_items || include_rowmargins)
			printf("\t%.8f", mpfr_get_d(blocks[i].node->cum_size, GMP_RNDN));
		printf("\n");
		if (include_rowmargins) {
			for (uint ii=0; ii<=item_cnt; ++ii)
				printf("%.4f ", mpfr_get_d(blocks[i].node->size_array[ii], GMP_RNDN));
			printf("\n");
			for (uint ii=0; ii<=item_cnt; ++ii)
				printf("%.4f ", mpfr_get_d(blocks[i].node->cum_size_array[ii], GMP_RNDN));
			printf("\n");
		}
	}
	
	printf("--------\n");
	
	return;
}
