
#include <mpfr.h>
#include "partitionnode.h"
#include "itemset.h"
#include "trie.h"

typedef unsigned int uint;



class ExpModel
{
public:
	
	ExpModel(uint n, uint d, bool use_items=false, bool use_rowmargins=false);
	~ExpModel();
	
	void add_itemset(int *itemset, int len, double freq, bool permanent=true);
	void set_item_frequency(int item, double frequency);
	void set_rowsize_prob(int size, double prob);
	
	double estimate_frequency(int *itemset, int len);
	
	void   iterative_scaling(int maxloops=100);
	double get_entropy();
	
	void compute_all_cum_sizes();
	void adjust_all_cum_sizes(int item, double oldparam, double newparam);
	void adjust_cum_sizes(PartitionNode * blocknode, double oldparam, double newparam);
	
public:
	
	void refine_partition(bool permanent);
	void remove_last_itemset();
	void delete_itemset(int j);
	
	void compute_block_sizes();
	void compute_block_size_arrays();
	void compute_block_probs(bool permanent=false);
	
	void compute_itemset_arrays(int i);
	
	void prune_blocks_if_empty();
	void prune_blocks_if_prob_zero();
	
	void interleave_blocks(uint offset, uint len);
	void reverse_blockrange(uint from, uint to);
	
	void qie(uint j, uint low, uint high);
	void qie_size_array(uint j, uint low, uint high);
	uint prefix_closure(const uint vect, const uint j, const PartitionNode *bloxnode);
	
	void normalize();
	
	double adjust_item_param(int item);
	
	void sync_params();
	
	uint get_index(uint from, uint to, uint t) const;
	uint get_first_mask_index(uint from, uint to, uint j) const;
	
	void print_model();

public:
	const uint item_cnt;
	const uint trans_cnt;
	const bool include_items;
	const bool include_rowmargins;
	
	PartitionBlock *blocks;
	uint inited_blocks;
	uint partition_size;
	
	Itemset **itemsets;
	double *frequencies;
	BitVector **itemsets_bits;
	uint itemsets_count;
	uint * itemsets_union;
	uint itemsets_union_size;
	BitVector *itemsets_union_bits;
	
	Itemset * dummy_itemset_adder;
	
	double *item_frequencies;
	double *item_params;
	double *rowsize_probs;
	mpfr_t *rowsize_estimates;
	
	// maxent model parameters
	mpfr_t  u0_param;
	mpfr_t  v0_param;
	mpfr_t  w0_param;
	mpfr_t *u_params;
	mpfr_t *v_params;
	mpfr_t *w_params;
	
	// temporary variables, initialized once
	mpfr_t estimate;
	mpfr_t sum;
	mpfr_t factor;
	mpfr_t h;
	mpfr_t f;
	mpfr_t f1, f2;
	mpfr_t tmp;
	
	BitVector *bitvect; // for closure
	
	static const unsigned int max_k = 8 * sizeof(unsigned int) - 2;
	static const unsigned int max_partition_size = (unsigned int)1 << 20;
	
	static const bool DO_QIE = true;

};




inline
void ExpModel::reverse_blockrange(uint from, uint to)
{
	uint r = (to - from) >> 1;
	
	for (uint i=0; i<r; ++i) {
		swap(blocks[from+i].index, blocks[to-i-1].index);
		swap(blocks[from+i].node, blocks[to-i-1].node);
	}
	
	return;
}

/*
 find first block index t in [from, to) for which t & mask != 0
 */
inline
uint ExpModel::get_first_mask_index(uint lo, uint hi, uint mask) const
{
	uint mid;
	
	while (lo < hi) {
		//mid = lo + ((hi-lo) >> 1);
		mid = (lo + hi) >> 1;
		if (blocks[mid].index & mask)
			hi = mid;
		else
			lo = mid + 1;
	}
	
	return hi;
}

/*
 find index t in [from, to)
 */
inline
uint ExpModel::get_index(uint from, uint to, uint t) const
{
	uint lo = from, hi = to, mid;
	
	while (lo < hi) {
		//mid = lo + ((hi-lo) >> 1);
		mid = (lo + hi) >> 1;
		if (blocks[mid].index < t)
			lo = mid + 1;
		else if (blocks[mid].index > t)
			hi = mid;
		else
			return mid;
	}
	
	return to;
}
