#include "expmodelgraph.h"
#include <math.h>
#include <set>
#include "auxiliary.h"

using std::set;


ExpModelGraph::ExpModelGraph(uint n, uint d, bool use_items, bool use_rowmargins) : item_cnt(n), trans_cnt(d), include_items(use_items), include_rowmargins(use_rowmargins)
{
	nodes = new ExpModel*[max_k+1];
	nodes[0] = 0;
	node_count = 0;
	
	all_itemsets = new Itemset*[max_k];
	frequencies = new double[max_k];
	all_itemsets_size = 0;
	
	item_frequencies = new double[item_cnt];
	for (uint i=0; i<item_cnt; ++i)
		item_frequencies[i] = 0.5;
	
	if (include_rowmargins) {
		rowsize_probs = new double[item_cnt+1];
		nodes[1] = new ExpModel(item_cnt, trans_cnt, include_items, include_rowmargins);
		node_count = 1;
	}
	else {
		rowsize_probs      = 0;
	}
	
	item_nodes = new uint[item_cnt];
	for (uint i=0; i<item_cnt; ++i)
		item_nodes[i] = 0;
	
	mpfr_init(estimate);
	mpfr_init(f);
	mpfr_init(h);
	
	projectedset = new int[n];
	thegroups = new uint[max_k];
	
	return;
}


ExpModelGraph::~ExpModelGraph()
{
	for (uint i=1; i<=node_count; ++i)
		delete nodes[i];
	delete [] nodes;
	
	for (uint i=0; i<all_itemsets_size; ++i)
		delete all_itemsets[i];
	delete [] all_itemsets;
	delete [] frequencies;
	delete [] item_frequencies;
	delete [] rowsize_probs;
	delete [] item_nodes;
	
	mpfr_clear(estimate);
	mpfr_clear(f);
	mpfr_clear(h);
	
	delete [] projectedset;
	delete [] thegroups;
	
	return;
}


void ExpModelGraph::add_itemset(int *itemset, int len, double freq)
{
	uint thenode = 0;
	
	if (include_rowmargins) {
		thenode = 1;
	}
	else {
		set<uint> groups;
		for (int i=0; i<len; ++i) {
			if (item_nodes[itemset[i]] != 0) {
				groups.insert(item_nodes[itemset[i]]);
				if (thenode)
					thenode = min(item_nodes[itemset[i]], thenode);
				else
					thenode = item_nodes[itemset[i]];
			}
		}
		
		if (thenode == 0) {
			// make new node
			++node_count;
			thenode = node_count;
			nodes[thenode] = new ExpModel(item_cnt, trans_cnt, include_items, include_rowmargins);
			for (uint item=0; item<item_cnt; ++item) {
				nodes[thenode]->set_item_frequency(item, item_frequencies[item]);
			}
		}
		else {
			// merges nodes
			groups.erase(thenode);
			for (set<uint>::iterator it=groups.begin(); it!=groups.end(); ++it) {
				for (uint i=0; i<nodes[*it]->itemsets_count; ++i) {
					nodes[thenode]->add_itemset(nodes[*it]->itemsets[i]->itemset, nodes[*it]->itemsets[i]->len, nodes[*it]->frequencies[i], true);
					for (int j=0; j<nodes[*it]->itemsets[i]->len; ++j)
						item_nodes[nodes[*it]->itemsets[i]->itemset[j]] = thenode;
				}
				delete nodes[*it];
				nodes[*it] = 0;
				nodes[thenode]->iterative_scaling();
			}
		}
	}
	
	nodes[thenode]->add_itemset(itemset, len, freq, true);
	
	all_itemsets[all_itemsets_size] = new Itemset(itemset, len);
	frequencies[all_itemsets_size] = freq;
	sort(all_itemsets[all_itemsets_size]->itemset, all_itemsets[all_itemsets_size]->itemset+len);
	++all_itemsets_size;
	
	for (int i=0; i<len; ++i)
		item_nodes[itemset[i]] = thenode;
	
	return;
}


void ExpModelGraph::set_item_frequency(int item, double frequency)
{
	item_frequencies[item] = frequency;
	
	for (uint i=1; i<=node_count; ++i) {
		if (nodes[i])
			nodes[i]->set_item_frequency(item, frequency);
	}
	
	return;
}


void ExpModelGraph::set_rowsize_prob(uint size, double prob)
{
	if (include_rowmargins) {
		for (uint i=1; i<=node_count; ++i)
			nodes[i]->set_rowsize_prob(size, prob);
		rowsize_probs[size] = prob;
	}
	
	return;
}


double ExpModelGraph::estimate_frequency(int *itemset, int len)
{
	if (include_rowmargins) {
		return nodes[1]->estimate_frequency(itemset, len);
	}
	else {
		mpfr_set_ui(estimate, 1, GMP_RNDN);
		
		groupsize = 0;
		for (int i=0; i<len; ++i) {
			if (item_nodes[itemset[i]]==0) {
				mpfr_mul_d(estimate, estimate, item_frequencies[itemset[i]], GMP_RNDN);
			}
			else if (find(thegroups, thegroups+groupsize, item_nodes[itemset[i]]) == thegroups+groupsize ) {
				thegroups[groupsize++] = item_nodes[itemset[i]];
				sort(thegroups, thegroups+groupsize);
			}
		}
		
		for (uint g = 0; g<groupsize; ++g) {
			int projlen = 0;
			for (int i=0; i<len; ++i) {
				if (item_nodes[itemset[i]] == thegroups[g]) {
					projectedset[projlen] = itemset[i];
					++projlen;
				}
			}
			mpfr_mul_d(estimate, estimate, nodes[thegroups[g]]->estimate_frequency(projectedset, projlen), GMP_RNDN);
		}
		
		return mpfr_get_d(estimate, GMP_RNDN);
	}
	
}


void ExpModelGraph::iterative_scaling(int maxloops)
{
	for (uint i=1; i<=node_count; ++i) {
		// TODO: only do iterative scaling when necessary
		if (nodes[i])
			nodes[i]->iterative_scaling(maxloops);
	}
	
	return;
}


void ExpModelGraph::delete_itemset(int j)
{
	// TODO: 
	
	// find in which node it is
	
	// remove it from that node
	
	// call iter scale on that node
	
	return;
}


double ExpModelGraph::get_entropy()
{
	mpfr_set_d(h, 0.0, GMP_RNDN);
	for (uint i=1; i<=node_count; ++i) {
		if (nodes[i])
			mpfr_add_d(h, h, nodes[i]->get_entropy(), GMP_RNDN);
	}
	
	if (!include_rowmargins) {
		for (uint i=0; i<item_cnt; ++i) {
			if (item_nodes[i] == 0) {
				double r = item_frequencies[i];
				if (r < 1.0 /*&& r > 0.0*/)
					mpfr_add_d(h, h, -(r * log_2(r) + (1-r) * log_2(1-r)), GMP_RNDN);
			}
		}
	}
	
	return mpfr_get_d(h, GMP_RNDN);
}


double ExpModelGraph::bg_entropy()
{
	mpfr_set_d(h, 0.0, GMP_RNDN);
	if (include_items) {
		if (include_rowmargins) {
			// TODO
		}
		else {
			for (uint i=0; i<item_cnt; ++i) {
				double & r = item_frequencies[i];
				if (r==0.0 || r==1.0)
					continue;
				mpfr_add_d(h, h, -(r * log_2(r) + (1.0-r)*log_2(1.0-r)), GMP_RNDN);
			}
		}
	}
	else {
		if (include_rowmargins) {
			mpfr_set_ui(f, 1, GMP_RNDN);
			mpfr_set_d(tmp, rowsize_probs[0], GMP_RNDN);
			mpfr_log2(tmp, tmp, GMP_RNDN);
			mpfr_mul_d(tmp, tmp, rowsize_probs[0], GMP_RNDN);
			mpfr_sub(h, h, tmp, GMP_RNDN);
			for (uint i=1; i<=item_cnt; ++i) {
				mpfr_mul_ui(f, f, item_cnt-i+1, GMP_RNDN);
				mpfr_div_ui(f, f, i, GMP_RNDN);
				mpfr_d_div(tmp, rowsize_probs[i], f, GMP_RNDN);
				mpfr_log2(tmp, tmp, GMP_RNDN);
				mpfr_mul_d(tmp, tmp, rowsize_probs[i], GMP_RNDN);
				mpfr_sub(h, h, tmp, GMP_RNDN);
			}
		}
		else {
			return item_cnt;
		}
	}
	
	return mpfr_get_d(h, GMP_RNDN);
}


void ExpModelGraph::compute_background_params()
{
	if (include_items) {
		if (include_rowmargins) {
			;//TODO
		}
		else
			;//trivial
	}
	else {
		if (include_rowmargins) {
			;//TODO
		}
	}
	
	return;
}


void ExpModelGraph::print_graph()
{
	for (uint i=1; i<=node_count; ++i) {
		if (nodes[i])
			nodes[i]->print_model();
	}
	
	return;
}
