#include <vector>
#include <string>
#include <iostream>
#include <algorithm>
#include <stdlib.h>

using std::vector;
using std::string;
using std::ostream;
using std::cerr;
using std::endl;
using std::find;


template <class T> inline unsigned int index_of(const vector<T> &v, T k) {
	unsigned int index = find(v.begin(), v.end(), k) - v.begin();
	if (index==v.size())
		return -1;
	return index;
}



class Options 
{
public:
	Options();
	int parse_options (int argc, char * argv[]);
	void print_usage(ostream & outstream);
	
	void add_option_strings(int ID, const char *short_str, const char *long_str, const char *help);

private:
	int get_option_code(string opt);
		
public:
	enum OPTIONS {HELP=0, FILENAME, K, MINSUP, MAXSIZE, OUTPUT, ITEMS, NOITEMS,\
		ROWMARGINS, NOROWMARGINS, FLY, NOFLY, PENALTY, LOWSTOP, TIMELIMIT, SEED, DICTIONARY, \
		ESTIMATE, MAXTREEWIDTH, MAXGROUPSIZE, TOPK, VERBOSITY, QUIET, OPT_CNT};
	
	vector<string> short_opts, long_opts, help_opts;

	string  filename;
	string  seedfilename;
	string  dictionaryfilename;
	string  outfilename;
	string  estimateinfilename;
	string  estimateoutfilename;
	
	int     k;
	double  minsup;
	int     maxsize;
	int     timelimit;
	bool	items;
	bool    fly;
	int     penalty;
	bool    lowstop;
	bool    rowmargins;
	int     maxtreewidth;
	int     maxgroupsize;
	int     topk;
	int     verbosity;
	bool    help;
	
};

Options::Options()
{
	short_opts = vector<string>(OPT_CNT);
	long_opts = vector<string>(OPT_CNT);
	help_opts = vector<string>(OPT_CNT);
	add_option_strings(FILENAME, "-f", "--file", "data filename, obligatory");
	add_option_strings(K, "-k", "--k", "number of itemsets in summary [default=10]");
	add_option_strings(MINSUP, "-s", "--minsup", "minimum support threshold [default=0.25]");
	add_option_strings(MAXSIZE, "-m", "--maxsize", "maximum itemset size threshold [default=0]");
	add_option_strings(OUTPUT, "-o", "--output", "summary filename");
	add_option_strings(ITEMS, "-i", "--items", "use item probs as background info [default]");
	add_option_strings(NOITEMS, "-j", "--no-items", "do not use item probabilities");
	add_option_strings(ROWMARGINS, "-r", "--rowmargins", "use row margins as background info");
	add_option_strings(NOROWMARGINS, "-a", "--no-rowmargins", "do not use row margins [default]");
	add_option_strings(FLY, "-y", "--fly", "mine frequencies on the fly [default]");
	add_option_strings(NOFLY, "-x", "--no-fly", "mine and store all frequencies beforehand");
	add_option_strings(PENALTY, "-p", "--penalty", "type of penalty [1=BIC, 3=MDL; default=3]");
	add_option_strings(LOWSTOP, "-l", "--low-stop", "stop when score increases [default=false]");
	add_option_strings(TIMELIMIT, "-t", "--time", "stop after x seconds have elapsed [default=0]");
	add_option_strings(SEED, "-c", "--seed", "initialize summary with itemsets from a file");
	add_option_strings(DICTIONARY, "-d", "--dictionary", "dictionary file containing item names");
	add_option_strings(ESTIMATE, "-e", "--estimate", "estimate itemsets in file and save to file");
	add_option_strings(MAXTREEWIDTH, "-w", "--maxwidth", "maximum number of items per group [default=0]");
	add_option_strings(MAXGROUPSIZE, "-g", "--maxgroupsize", "maximum number of itemsets per group [default=0]");
	add_option_strings(TOPK, "-b", "--b", "at each iteration, show the top ? scoring itemsets [default=0]");
	add_option_strings(VERBOSITY, "-v", "--verbosity", "verbosity level [default=1]");
	add_option_strings(QUIET, "-q", "--quiet", "quiet mode [verbose=0]");
	add_option_strings(HELP, "-h", "--help", "display this help and exit");
	
	filename            = string("");
	seedfilename        = string("");
	dictionaryfilename  = string("");
	outfilename         = string("");
	estimateinfilename  = string("");
	estimateoutfilename = string("");
	k            = 10;
	items        = true;
	minsup       = 0.25;
	maxsize      = 0;
	timelimit    = 0;
	fly          = true;
	penalty      = 3;
	lowstop      = false;
	rowmargins   = false;
	maxtreewidth = 0;
	maxgroupsize = 0;
	topk         = 0;
	verbosity    = 1;
	help         = false;
	
	return;
}

void Options::add_option_strings(int ID, const char *short_str, const char *long_str, const char *help_str)
{
	short_opts[ID] = short_str;
	long_opts[ID]  = long_str;
	help_opts[ID]  = help_str;
	
	return;
}	

int Options::get_option_code(string opt)
{
	int option = index_of(short_opts, opt);
	if (option == -1)
		option = index_of(long_opts, opt);
	
	return option;
}


void Options::print_usage(ostream &outstream)
{
	outstream << "USAGE:" << endl;
	for (int i=0; i<OPT_CNT; ++i) {
		outstream << short_opts[i];
		outstream << '\t';
		outstream << long_opts[i];
		for (int tt=0; tt < 3 - int(long_opts[i].size())/8; ++tt)
			outstream << '\t';
		outstream << help_opts[i];
		outstream << endl;
	}
	
	return;
}


int Options::parse_options (int argc, char * argv[])
{
	for (int ar=1; ar<argc; ++ar) {
		int option = get_option_code(string(argv[ar]));
		
		switch (option) {
			case HELP:
				help = true;
				break;
			case FILENAME:
				filename = string(argv[ar+1]);
				++ar;
				break;
			case K:
				k = atoi(argv[ar+1]);
				++ar;
				break;
			case OUTPUT:
				outfilename = string(argv[ar+1]);
				++ar;
				break;
			case MINSUP:
				minsup = atof(argv[ar+1]);
				++ar;
				break;
			case MAXSIZE:
				maxsize = atoi(argv[ar+1]);
				++ar;
				break;
			case ITEMS:
				items = true;
				break;
			case NOITEMS:
				items = false;
				break;
			case ROWMARGINS:
				rowmargins = true;
				break;
			case NOROWMARGINS:
				rowmargins = false;
				break;
			case FLY:
				fly = true;
				break;
			case NOFLY:
				fly = false;
				break;
			case PENALTY:
				penalty = atoi(argv[ar+1]);
				++ar;
				break;
			case LOWSTOP:
				lowstop = true;
				break;
			case TIMELIMIT:
				timelimit = atoi(argv[ar+1]);
				++ar;
				break;
			case SEED:
				seedfilename = string(argv[ar+1]);
				++ar;
				break;
			case DICTIONARY:
				dictionaryfilename = string(argv[ar+1]);
				++ar;
				break;
			case ESTIMATE:
				estimateinfilename = string(argv[ar+1]);
				estimateoutfilename = string(argv[ar+2]);
				ar += 2;
				break;
			case MAXTREEWIDTH:
				maxtreewidth = atoi(argv[ar+1]);
				++ar;
				break;
			case MAXGROUPSIZE:
				maxgroupsize = atoi(argv[ar+1]);
				++ar;
				break;
			case TOPK:
				topk = atoi(argv[ar+1]);
				++ar;
				break;
			case VERBOSITY:
				verbosity = atoi(argv[ar+1]);
				++ar;
				break;
			case QUIET:
				verbosity = 0;
				break;
			default:
				if (string(argv[ar]).operator [](0) == '-') {
					cerr << "Unknown option: " << argv[ar] << endl;
					return 1;
				}
				break;
		}
	}
	
	return 0;
}

