#include <iostream>
#include "summaryminer.h"
#include <stdlib.h>
#include "options.h"
#include <time.h>

using namespace std;


const string version = "v2.0.20111208";
const string author  = "Michael Mampaey";
const string url     = "http://www.adrem.ua.ac.be";


int main (int argc, char * argv[])
{
	Options options;
	int parsed = options.parse_options(argc, argv);
	
	if (parsed != 0)
		exit(1);
	
	cout.precision(8);
	
	int verbosity = options.verbosity;
	
	if (verbosity >= 1)
		cout << "Summarization with itemsets " << version << endl << url << endl << endl;
	
	if (argc == 1 || options.help) {
		options.print_usage(cerr);
		return 1;
	}
	
	if (options.filename==string("")) {
		cerr << "No filename given" << endl;
		return 1;
	}
	
	clock_t before_read, after_read;
	clock_t before_mine, after_mine;
	
	SummaryMiner sm(verbosity, options.topk);
	
	sm.set_use_items(options.items);
	sm.set_use_rowmargins(options.rowmargins);
	
	if (verbosity >= 1)
		cout << "Reading data file: '" << options.filename << "' ... " << flush;
	
	before_read = clock();
	sm.read_data(options.filename);
	after_read = clock();
	
	if (verbosity >= 1)
		cout << sm.item_count << " items" << ' ' << sm.trans_count << " transactions" << endl;
	
	if (options.seedfilename.size()) {
		if (verbosity >= 1)
			cout << "Seeding summary with: '" << options.seedfilename << "' ... " << flush;
		int seeds = sm.seed_summary(options.seedfilename);
		if (verbosity >= 1)
			cout << seeds << " sets added" << endl;	
	}
	
	sm.set_minsup(options.minsup);
	sm.set_maxsize(options.maxsize);
	sm.set_maxtreewidth(options.maxtreewidth);
	sm.set_maxgroupsize(options.maxgroupsize);
	sm.set_penalty(options.penalty);
	
	if (options.outfilename.size() != 0 || verbosity >= 1)
		sm.read_dictionary(options.dictionaryfilename);
	
	if (verbosity >= 1)
		cout << "Mining summary of size: " << options.k << endl;
	
	before_mine = clock();
	sm.mine_summary(options.k, options.lowstop, options.timelimit, options.fly);
	after_mine = clock();
	
	if (verbosity >= 1)
		cout << "Done" << endl << endl;
	
	if (verbosity >= 2) {
		double bg_entropy = 0.0;
		if (options.rowmargins || options.items)
			bg_entropy = sm.bg_entropy;
		else
			bg_entropy= sm.summary->bg_entropy();
		double bg_penalty = sm.get_penalty(0, 0);
		double entropy = sm.summary->get_entropy();
		double penalty = sm.get_penalty();
		double data_entropy = sm.data_entropy();
		int entropy_width = log(sm.summary->item_cnt) / log(10.0) + 8;
		int score_width = log(sm.summary->item_cnt * sm.trans_count) / log(10.0) + 4;
		printf("\t\tENTROPY\t\t\tSCORE\n");
		printf("  uniform     : %*.6f\t\t%*.2f\n", entropy_width, double(sm.summary->item_cnt), score_width, sm.summary->item_cnt * sm.trans_count + bg_penalty);
		if (options.items || options.rowmargins)
			printf("  background  : %*.6f\t\t%*.2f\n", entropy_width, bg_entropy, score_width, bg_entropy * sm.trans_count + bg_penalty);
		printf("  summary     : %*.6f\t\t%*.2f\n", entropy_width, entropy, score_width, entropy * sm.trans_count + penalty);
		printf("  data        :<%*.6f\t\t-\n", entropy_width, data_entropy);
	}
	
	
	if (options.outfilename.size())
		sm.save_summary(options.outfilename);
	
	if (verbosity >= 1) {
		cout << endl << "SUMMARY:" << endl;
		sm.print_summary();
	}
	
	if (options.estimateinfilename.size() != 0) {
		if (verbosity >= 1)
			cout << "Estimating itemset frequencies in '" << options.estimateinfilename << "' ... "<< flush;
		sm.estimate_itemsets(options.estimateinfilename, options.estimateoutfilename);
		if (verbosity >= 1)
			cout << "done" << endl;
	}
	
	if (verbosity >= 1) {
		cout << endl << "TIME:" << endl;
		cout << "  read:\t" << (after_read-before_read) / CLOCKS_PER_SEC << " s" << endl;
		cout << "  mine:\t" << (after_mine-before_mine) / CLOCKS_PER_SEC << " s" << endl;
	}
	
	return 0;
	
}
