#include "summaryminer.h"
#include <vector>
#include <stdio.h>
#include <set>
#include <map>
#include <algorithm>
#include <cmath>
#include <iostream>
#include <fstream>
#include <limits>
#include <exception>
#include <time.h>
#include "auxiliary.h"

using namespace std;



struct TidListComp
{
	bool operator()(const TidList* one, const TidList *two) {
		return one->size < two->size;
	}
} tidlistcomp;


struct ItemFrequencyComp
{
	bool operator()(const ItemFrequency & one, const ItemFrequency & two) {
		return one.frequency < two.frequency;
	}
} itemfrequencycomp;



SummaryMiner::SummaryMiner(int verbose, int topk)
{
	top_itemset = Itemset();
	top_score = 0.0;
	top_freq  = 0.0;
	top_est   = 0.0;
	
	top_k = topk;
	do_top_k_sets = (top_k > 0);
	if (do_top_k_sets) {
		top_k_itemsets = new Itemset [top_k];
		top_k_scores = new double[top_k];
		top_k_freqs  = new double[top_k];
		top_k_ests   = new double[top_k];
		for (int i=0; i<top_k; ++i) {
			top_k_itemsets[i] = Itemset();
			top_k_scores[i] = 0.0;
			top_k_freqs[i]  = 0.0;
			top_k_ests[i]   = 0.0;
		}
	}
	
	datafilename = string("");
	item_count  = 0;
	trans_count = 0.0;
	freq_item_count = 0;
	
	minsup    = 0.0;
	absminsup = 0.0;
	
	maxtreewidth = 0;
	maxgroupsize = 0;
	
	summary  = 0;
	tidlists = 0;
	
	use_items      = true;
	use_rowmargins = false;
	
	rowsize_probs = 0;
	
	freq_trie = 0;
	
	dict = 0;
	
	verbosity = verbose;
	
	return;
}


SummaryMiner::~SummaryMiner()
{
	delete summary;
	for (int i=0; i<freq_item_count; ++i) {
		delete tidlists[i];
	}
	delete [] tidlists;
	delete freq_trie;
	
	TidList::unprepare();
	
	delete [] rowsize_probs;
	
	delete [] dict;
	
	delete [] id2item;
	
	if (do_top_k_sets) {
		delete [] top_k_itemsets;
		delete [] top_k_scores;
		delete [] top_k_freqs;
		delete [] top_k_ests;
	}
	
	return;
}


void SummaryMiner::init_summary()
{
	static bool initialized = false;
	
	if (!initialized) {
		summary = new ExpModelGraph(item_count, trans_count, use_items, use_rowmargins);
		initialized = true;
	}
	
	return;
}


void SummaryMiner::read_data(string filename)
{
	datafilename = filename;
	
	ifstream *datafile = new ifstream(datafilename.c_str());
	if (datafile == 0) {
		cerr << endl << "Error: Data file '" << datafilename << "' does not exist."<< endl;
		throw(exception());
	}
	
	trans_count = 0.0;
	int item;
	uint pos = 0;
	
	map<int, int> rowsizes;
	
	map<int, vector<int> > tmptids;
	string line;
	while (!datafile->eof()) {
		getline(*datafile, line);
		if (line.size() == 0) {
			if (!datafile->eof())
				++trans_count;
		}
		else if (line[0] == '#') {
			continue;
		}
		else {
			int tsize = 0;
			pos = 0;
			while (pos < line.size()) {
				bool pres = false;
				item = 0;
				while (pos < line.size() && line[pos] >= '0' && line[pos] <= '9') {
					pres = true;
					item *= 10;
					item += line[pos] - '0';
					++pos;
				}
				if (pres) {
					if (tmptids.find(item) == tmptids.end()) {
						tmptids[item] = vector<int>();
					}
					tmptids[item].push_back((int)trans_count);
					++tsize;
				}
				++pos;
			}
			if (rowsizes.find(tsize) == rowsizes.end())
				rowsizes[tsize] = 0;
			rowsizes[tsize] += 1;
			++trans_count;
		}
	}
	
	datafile->close();
	
	item_count = tmptids.size();
	freq_item_count = item_count;
	tidlists = new TidList*[item_count];
	item2id.clear();
	id2item = new int[item_count];
	
	int item_id = 0;
	for (map<int,vector<int> >::iterator it=tmptids.begin(); it!=tmptids.end(); ++it) {
		item2id[it->first] = item_id;
		id2item[item_id] = it->first;
		tidlists[item_id] = new TidList(item_id, it->second);
		++item_id;
	}
	
	if (use_rowmargins) {
		rowsize_probs = new double[item_count+1];
		for (uint i=0; i<=item_count; ++i) {
			if (rowsizes.find(i) == rowsizes.end())
				rowsize_probs[i] = 0.0;
			else
				rowsize_probs[i] = rowsizes[i] / trans_count;
		}
	}
	
	TidList::prepare((int)trans_count);
	
	absminsup = minsup * trans_count;
	
	return;
}


double SummaryMiner::data_entropy()
{
	
	return log_2(trans_count); // TODO: this is an upper bound
}


int SummaryMiner::seed_summary(string filename)
{
	
	init_summary();
	
	// TODO: set item freqs before seeding
	
	for (uint i=0; i<item_count; ++i) {
		if (use_items || use_rowmargins)
			summary->set_item_frequency(i, tidlists[i]->size/trans_count);
	}
	
	ifstream *seedfile = new ifstream(filename.c_str());
	if (seedfile == 0) {
		cerr << endl << "Error: Seed file '" << datafilename << "' does not exist."<< endl;
		throw(exception());
	}
	
	int *tmpitemset = new int[item_count];
	
	uint pos = 0;
	int len = 0;
	double freq= 0.0;
	string line;
	while (!seedfile->eof()) {
		getline(*seedfile, line);
		
		if (line.size()==0 || line[0] == '#') {
			continue;
		}
		else {
			pos = 0;
			while (pos < line.size() && line[pos] != ' ' && line[pos] != '\t')
				++pos;
			freq = atof(line.substr(0, pos).c_str());
			++pos;
			
			int item;
			len = 0;
			while (pos < line.size()) {
				bool pres = false;
				item = 0;
				while (pos < line.size() && line[pos] >= '0' && line[pos] <= '9') {
					pres = true;
					item *= 10; 
					item += line[pos] - '0';
					++pos;
				}
				if (pres)
					tmpitemset[len++] = item2id[item];
				++pos;
			}
			
			summary->add_itemset(tmpitemset, len, freq);
			summary->iterative_scaling();
		}
		
	}
	
	delete [] tmpitemset;
	seedfile->close();
	
	if (verbosity >= 2) {
		printf("Entropy: %g\n", summary->get_entropy());
		if (verbosity >= 3)
			summary->print_graph();
	}
	
	return summary->all_itemsets_size;
}


void SummaryMiner::mine_summary(int l, bool lowstop, int timelimit, bool onthefly)
{
	time_t start_time, current_time;
	start_time = time(0);
	
	init_summary();
	
	int rm_cnt = 0;
	for (uint i=0; i<item_count; ++i) {
		if (use_items||use_rowmargins)
			summary->set_item_frequency(i, tidlists[i]->size/trans_count); // TODO: set also before seeding
		if (tidlists[i]->size < absminsup || tidlists[i]->size==(int)trans_count) {
			delete tidlists[i];
			++rm_cnt;
		}
		else {
			tidlists[i-rm_cnt] = tidlists[i];
		}
	}
	freq_item_count -= rm_cnt;
	
	if (use_rowmargins) {
		for (uint i=0; i<=item_count; ++i)
			summary->set_rowsize_prob(i, rowsize_probs[i]);
		
	}
	
	if (maxsize == 0)
		maxsize = freq_item_count;
	
	prefix  = new int[freq_item_count];
	closure = new int[freq_item_count];
	prefix_nodes  = new int[freq_item_count];
	closure_nodes = new int[freq_item_count];
	
	ItemFrequency *items = 0;
	if (!onthefly) {
		if (verbosity >= 1) {
			printf("Mining and storing all frequent sets off the fly ... ");
			fflush(stdout);
		}
		freq_trie = new Trie<double>(-1.0);
		mine_and_store_frequent_sets(tidlists, freq_item_count);
		items = new ItemFrequency[freq_item_count];
		for (int i=0; i<freq_item_count; ++i) {
			items[i].item      = tidlists[i]->id;
			items[i].frequency = tidlists[i]->size / trans_count;
		}
		if (verbosity >= 1)
			printf("%i sets.\n", freq_trie->get_trie_size());
	}
	
	if (verbosity >= 1) {
		// header
		if (penalty_type==1)
			printf("Nr\theuristic\t\t      BIC\t\t time\t\tcand\n");
		else if (penalty_type==2)
			printf("Nr\theuristic\t\t      MDL\t\t time\t\tcand\n");
		else if (penalty_type==3)
			printf("Nr\theuristic\t\t      MDL\t\t time\t\tcand\n");
	}
	
	//summary->print_graph();
	
	summary->iterative_scaling();
	bg_entropy = summary->get_entropy();
	
	double prev_score = bg_entropy * trans_count + get_penalty(0,0);
	
	int k = summary->all_itemsets_size;
	while (k++ < l)
	{
		if (verbosity >= 1) {
			printf("%2i\t", k);
			fflush(stdout);
		}
		
		top_itemset = Itemset();
		top_score   = (penalty_type==3) ? - (item_count*get_mdl_penalty_2_delta() / trans_count) : 0.0;
		top_freq    = 0.0;
		top_est     = 0.0;
		if (do_top_k_sets) {
			for (int i=0; i<top_k; ++i) {
				top_k_itemsets[i] = Itemset();
				top_k_scores[i]   = (penalty_type==3) ? - (item_count*get_mdl_penalty_2_delta() / trans_count) : 0.0;
				top_k_freqs[i]    = 0.0;
				top_k_ests[i]     = 0.0;
			}
		}
		
		uint cand;
		if (onthefly)
			cand = mine_best_itemset(tidlists, freq_item_count);
		else
			cand = mine_best_itemset_offthefly(items, freq_item_count);
		if (verbosity >= 1) {
			printf("%.7f", top_score);
			fflush(stdout);
		}
		
		if (top_score == ((penalty_type==3) ? - (item_count*get_mdl_penalty_2_delta() / trans_count) : 0.0)) // smaller than some epsilon for rounding errors?
			break;
		
		summary->add_itemset(top_itemset.itemset, top_itemset.len, top_freq);
		summary->iterative_scaling();
		
		double entropy = summary->get_entropy();
		double penalty = get_penalty();
		double score = entropy * trans_count + penalty;
		if (lowstop && score >= prev_score) {
			if (verbosity >= 1)
				printf("\n");
			break;
		}
		
		current_time = time(0);
		
		if (verbosity >= 2) {
			printf("\t\t%.2f\t\t%4is\t\t%u\n", score, (int)difftime(current_time, start_time), cand);
		}
		else if (verbosity >= 1)
			printf("\n");
		if (verbosity >= 3)
			summary->print_graph();
		
		if (do_top_k_sets && verbosity >= 1)
			print_top_k();
		
		if (timelimit && difftime(current_time, start_time) > timelimit) {
			if (verbosity >= 1)
				printf("\n");
			break;
		}		
		
		prev_score = score;
	}
	if (verbosity >= 1)
		printf("\n");
	
	delete [] prefix;
	prefix = 0;
	delete [] closure;
	closure = 0;
	delete [] prefix_nodes;
	prefix_nodes = 0;
	delete [] closure_nodes;
	closure_nodes = 0;
	delete [] items;
	items = 0;
	
	return;
}


void SummaryMiner::mine_and_store_frequent_sets(TidList **items, int tailsize, int depth)
{
	sort(items, items+tailsize, tidlistcomp);
	
	for (int i=0; i<tailsize; ++i) {
		prefix[depth] = items[i]->id;
		
		freq_trie->add(prefix, depth+1, items[i]->size / trans_count);
		
		if (depth+1 >= maxsize) continue;
		
		TidList **children = new TidList*[tailsize];
		
		int tailsize2 = 0;
		for (int j=i+1; j<tailsize; ++j) {
			TidList *item;
			if (depth == 0)
				item = new TidList(items[j]->id, *items[i], *items[j]);
			else
				item = new TidList(items[j]->id, *items[j], *items[i]);
			item->size = items[i]->size - item->len;
			
			if (item->size >= absminsup)
				children[tailsize2++] = item;
			else
				delete item;
		}
		
		mine_and_store_frequent_sets(children, tailsize2, depth + 1);
		
		for (int j=0; j<tailsize2; ++j)
			delete children[j];
		delete [] children;
		
	}
	
	return;
}


uint SummaryMiner::mine_best_itemset(TidList **items, int tailsize, int clen, int depth)
{
	sort(items, items+tailsize, tidlistcomp);
	
	uint cand = 0;
	
	for (int i=0; i<tailsize; ++i) {
		prefix[depth] = items[i]->id;
		if (maxtreewidth || maxgroupsize) { // check for the free sets first
			prefix_nodes[depth] = summary->item_nodes[items[i]->id];
			uint nodewidth = 0;
			uint nodegsize = 0;
			set<int> nodes = set<int>();
			for (int j=0; j<depth+1; ++j) {
				if (prefix_nodes[j])
					nodes.insert(prefix_nodes[j]);
				else
					++nodewidth;
			}
			for (set<int>::iterator it=nodes.begin(); it!=nodes.end(); ++it) {
				nodegsize += summary->nodes[*it]->itemsets_count;
				nodewidth += summary->nodes[*it]->itemsets_union_size;
			}
			if ((maxtreewidth && nodewidth > maxtreewidth) || (maxgroupsize && nodegsize+1 > maxgroupsize))
				continue;
		}
		
		++cand;
		
		double freq = items[i]->size / trans_count;
		
		// if itemset has top score, it is either a closed or free set
		// full closure is done below, since possibly clen2 > clen
		double est_freq_free  = summary->estimate_frequency(prefix, depth + 1); // what about the closure?
		double score_free     = heuristic(freq, est_freq_free) - (penalty_type==3 ? (depth+1)*get_mdl_penalty_2_delta()/trans_count : 0.0 );
		if (score_free > top_score) {
			top_itemset = Itemset(prefix, depth + 1);
			top_score   = score_free;
			top_freq    = freq;
			top_est     = est_freq_free;
			if (do_top_k_sets)
				insert_top_k_itemset(top_itemset, top_score, top_freq, top_est);
		}
		else if (do_top_k_sets && score_free > top_k_scores[top_k-1]) {
			Itemset tmpset = Itemset(prefix, depth + 1);
			insert_top_k_itemset(tmpset, score_free, freq, est_freq_free);
		}
		
		// checks 4-supp bound on free set
		double tail_freq = 0.0;
		if (depth == 0)
			tail_freq = TidList::multi_intersect_size(items[i], items+i+1, tailsize-i-1) / trans_count;
		else
			tail_freq = (items[i]->size - TidList::multi_union_diff_size(items[i], items+i+1, tailsize-i-1)) / trans_count;
		tail_freq = max(tail_freq, minsup);
		for (int j=i+1; j<tailsize; ++j)
			prefix[depth+j-i] = items[j]->id;
		for (int j=0; j<clen; ++j)
			prefix[depth+tailsize-i+j] = closure[j];
		double est_tail_freq  = summary->estimate_frequency(prefix, depth  + tailsize - i + clen);
		double score_bound = max(heuristic(freq, est_tail_freq), heuristic(est_freq_free, tail_freq)) - (penalty_type==3 ? (depth+2)*get_mdl_penalty_2_delta()/trans_count : 0.0 );;
		
		if (score_bound < top_score)
			continue;
		
		TidList **children = new TidList*[tailsize-i];
		int max_child_supp = 0;
		int clen2 = clen;
		int tailsize2 = 0;
		
		for (int j=i+1; j<tailsize; ++j) {
			TidList *item;
			if (depth==0)
				item = new TidList(items[j]->id, *items[i], *items[j]);
			else
				item = new TidList(items[j]->id, *items[j], *items[i]);
			item->size = items[i]->size - item->len;
			
			if (item->size == items[i]->size) { // item->len==0
				closure[clen2] = item->id;
				if (maxtreewidth || maxgroupsize)
					closure_nodes[clen2] = summary->item_nodes[item->id];
				++clen2;
				delete item;
				continue;
			}
			
			if (item->size >= absminsup) {
				children[tailsize2++] = item;
				if (item->size > max_child_supp)
					max_child_supp = item->size;
			}
			else
				delete item;
		}
		
		if (depth+1+clen2 > maxsize) {
			for (int j=0; j<tailsize2; ++j)
				delete children[j];
			delete [] children;
			continue;
		}
		
		if (max_child_supp < items[i]->size && clen2 == 0) {
			double score_bound = max(heuristic(max_child_supp/trans_count, est_tail_freq), heuristic(est_freq_free, tail_freq)); // 2nd == score
			score_bound -= penalty_type==3 ? (depth+2)*get_mdl_penalty_2_delta()/trans_count : 0.0 ;
			if (score_bound < top_score) {
				for (int j=0; j<tailsize2; ++j)
					delete children[j];
				delete [] children;
				continue;
			}
		}
		
		if (clen2) { // update top_score for full closure
			for (int j=0; j<clen2; ++j) 
				prefix[depth+1+j] = closure[j];
			double est_freq_cl  = summary->estimate_frequency(prefix, depth + 1 + clen2);
			double score_cl     = heuristic(freq, est_freq_cl) - (penalty_type==3 ? (depth+1+clen2)*get_mdl_penalty_2_delta()/trans_count : 0.0 );;
			if (score_cl > top_score) {
				if (maxtreewidth || maxgroupsize) { // check if we're allowed to by maxtreewidth
					uint nodewidth = 0;
					uint nodegsize = 0;
					set<int> nodes = set<int>();
					for (int j=0; j<depth+1; ++j) {
						if (prefix_nodes[j])
							nodes.insert(prefix_nodes[j]);
						else
							++nodewidth;
					}
					for (int j=0; j<clen2; ++j) {
						if (closure_nodes[j])
							nodes.insert(closure_nodes[j]);
						else
							++nodewidth;
					}
					for (set<int>::iterator it=nodes.begin(); it!=nodes.end(); ++it) {
						nodegsize += summary->nodes[*it]->itemsets_count;
						nodewidth += summary->nodes[*it]->itemsets_union_size;
					}
					if ((maxtreewidth && nodewidth > maxtreewidth) || (maxgroupsize && nodegsize+1 > maxgroupsize)) {
						for (int j=0; j<tailsize2; ++j)
							delete children[j];
						delete [] children;
						continue;
					}
				}
				
				top_itemset = Itemset(prefix, depth + 1 + clen2);
				top_score   = score_cl;
				top_freq    = freq;
				top_est     = est_freq_cl;
				if (do_top_k_sets)
					insert_top_k_itemset(top_itemset, top_score, top_freq, top_est);
				
				if (score_bound < top_score) {
					for (int j=0; j<tailsize2; ++j)
						delete children[j];
					delete [] children;
					continue;
				}
			}
			else if (do_top_k_sets && score_cl > top_k_scores[top_k-1]) {
				Itemset tmpset = Itemset(prefix, depth + 1 + clen2);
				insert_top_k_itemset(tmpset, score_cl, freq, est_freq_cl);
			}
		}
		
		cand += mine_best_itemset(children, tailsize2, clen2, depth + 1);
		
		for (int j=0; j<tailsize2; ++j)
			delete children[j];
		delete [] children;
		
	}
	
	return cand;
}


uint SummaryMiner::mine_best_itemset_offthefly(ItemFrequency *items, int tailsize, int clen, int depth)
{
	sort(items, items+tailsize, itemfrequencycomp);
	
	uint cand = 0;
	
	for (int i=0; i<tailsize; ++i) {
		prefix[depth] = items[i].item;
		
		if (maxtreewidth || maxgroupsize) { // check for the free sets first
			prefix_nodes[depth] = summary->item_nodes[items[i].item];
			uint nodewidth = 0;
			uint nodegsize = 0;
			set<int> nodes = set<int>();
			for (int j=0; j<depth+1; ++j) {
				if (prefix_nodes[j])
					nodes.insert(prefix_nodes[j]);
				else
					++nodewidth;
			}
			for (set<int>::iterator it=nodes.begin(); it!=nodes.end(); ++it) {
				nodegsize += summary->nodes[*it]->itemsets_count;
				nodewidth += summary->nodes[*it]->itemsets_union_size;
			}
			if ((maxtreewidth && nodewidth > maxtreewidth) || (maxgroupsize && nodegsize+1 > maxgroupsize))
				continue;
		}
		
		++cand;
		
		double freq = items[i].frequency;
		
		// if itemset has top score, it is either a closed or free set
		// full closure is done below, since possibly clen2 > clen
		double est_freq_free = summary->estimate_frequency(prefix, depth + 1); 
		double score_free    = heuristic(freq, est_freq_free) - (penalty_type==3 ? (depth+1)*get_mdl_penalty_2_delta()/trans_count : 0.0 );;
		if (score_free > top_score) {
			top_itemset = Itemset(prefix, depth + 1);
			top_score   = score_free;
			top_freq    = freq;
			top_est     = est_freq_free;
			if (do_top_k_sets)
				insert_top_k_itemset(top_itemset, top_score, top_freq, top_est);
		}
		else if (do_top_k_sets && score_free > top_k_scores[top_k-1]) {
			Itemset tmpset = Itemset(prefix, depth + 1);
			insert_top_k_itemset(tmpset, score_free, freq, est_freq_free);
		}
		
		// checks 4-supp bound on free set
		for (int j=i+1; j<tailsize; ++j)
			prefix[depth+j-i] = items[j].item;
		double tail_freq      = freq_trie->find(prefix, depth + tailsize - i);
		for (int j=0; j<clen; ++j) 
			prefix[depth  + tailsize - i + j] = closure[j];
		double est_tail_freq  = summary->estimate_frequency(prefix, depth  + tailsize - i + clen);
		tail_freq = (tail_freq == -1.0) ? minsup : tail_freq ;
		double score_bound = max(heuristic(freq, est_tail_freq), heuristic(est_freq_free, tail_freq));
		score_bound -= penalty_type==3 ? (depth+2)*get_mdl_penalty_2_delta()/trans_count : 0.0 ;
		
		if (score_bound < top_score)
			continue;
		
		ItemFrequency *children = new ItemFrequency[tailsize-i];
		int clen2 = clen;
		int tailsize2 = 0;
		double max_child_freq = 0;
		
		for (int j=i+1; j<tailsize; ++j) {
			int item = items[j].item;
			prefix[depth+1] = item;
			double cfreq = freq_trie->find(prefix, depth+2);
			if (cfreq != -1.0) {
				if (cfreq == freq) {
					closure[clen2] = item;
					closure_nodes[clen2] = summary->item_nodes[item];
					++clen2;
					continue;
				}
				children[tailsize2].item = item;
				children[tailsize2].frequency = cfreq;
				++tailsize2;
				if (cfreq > max_child_freq)
					max_child_freq = cfreq;
			}
		}
		
		if (depth+1+clen2 > maxsize) {
			delete [] children;
			continue;
		}
		
		if (max_child_freq < freq && clen2 == 0) {
			double score_bound = max(heuristic(max_child_freq, est_tail_freq), heuristic(est_freq_free, tail_freq)); // 2nd == score
			score_bound -= penalty_type==3 ? (depth+1)*get_mdl_penalty_2_delta()/trans_count : 0.0 ;
			if (score_bound < top_score) {
				delete [] children;
				continue;
			}
		}
		
		if (clen2) { // update top_score for full closure
			for (int j=0; j<clen2; ++j) 
				prefix[depth+1+j] = closure[j];
			double est_freq_cl  = summary->estimate_frequency(prefix, depth + 1 + clen2);
			double score_cl     = heuristic(freq, est_freq_cl) - (penalty_type==3 ? (depth+1+clen2)*get_mdl_penalty_2_delta()/trans_count : 0.0 );
			if (score_cl > top_score) {
				if (maxtreewidth || maxgroupsize) { // check if we're allowed to by maxtreewidth
					uint nodewidth = 0;
					uint nodegsize = 0;
					set<int> nodes = set<int>();
					for (int j=0; j<depth+1; ++j) {
						if (prefix_nodes[j])
							nodes.insert(prefix_nodes[j]);
						else
							++nodewidth;
					}
					for (int j=0; j<clen2; ++j) {
						if (closure_nodes[j])
							nodes.insert(closure_nodes[j]);
						else
							++nodewidth;
					}
					for (set<int>::iterator it=nodes.begin(); it!=nodes.end(); ++it) {
						nodegsize += summary->nodes[*it]->itemsets_count;
						nodewidth += summary->nodes[*it]->itemsets_union_size;
					}
					if ((maxtreewidth && nodewidth > maxtreewidth) || (maxgroupsize && nodegsize+1 > maxgroupsize)) {
						delete [] children;
						continue;
					}
				}
				top_itemset = Itemset(prefix, depth + 1 + clen2);
				top_score   = score_cl;
				top_freq    = freq;
				top_est     = est_freq_cl;
				if (do_top_k_sets)
					insert_top_k_itemset(top_itemset, top_score, top_freq, top_est);
				
				if (score_bound < top_score) {
					delete [] children;
					continue;
				}
			}
			else if (do_top_k_sets && score_cl > top_k_scores[top_k-1]) {
				Itemset tmpset = Itemset(prefix, depth + 1 + clen2);
				insert_top_k_itemset(tmpset, score_cl, freq, est_freq_cl);
			}
		}
		
		
		cand += mine_best_itemset_offthefly(children, tailsize2, clen2, depth + 1);
		
		delete [] children;
		
	}
	
	return cand;
}


void SummaryMiner::estimate_itemsets(string est_in_filename, string est_out_filename)
{
	ifstream *estinfile = new ifstream(est_in_filename.c_str());
	if (estinfile == 0) {
		cerr << endl << "Error: Itemset file '" << est_in_filename << "' does not exist."<< endl;
		throw(exception());
	}
	
	FILE *estoutfile = fopen(est_out_filename.c_str(), "w");
	
	int *itemset = new int[100];
	int len = 0;
	
	int item = 0;
	uint pos;
	string line;
	int abssup = -1; // TODO: read support
	
	while (!estinfile->eof()) {
		getline(*estinfile, line);
		if (line.size() == 0) {
			if (!estinfile->eof())
				fprintf(estoutfile, " 1.0\n");
		}
		else if (line[0] == '#') {
			continue;
		}
		else {
			pos = 0;
			len = 0;
			while (pos < line.size()) {
				bool pres = false;
				item = 0;
				while (pos < line.size() && line[pos] >= '0' && line[pos]<='9') {
					pres = true;
					item *= 10;
					item += line[pos] - '0';
					++pos;
				}
				if (pres) {
					itemset[len++] = item2id[item];
				}
				if (line[pos] == '(') {
					++pos;
					abssup = 0;
					while (pos < line.size() && line[pos] != ')') {
						abssup *= 10;
						abssup += line[pos] - '0';
						++pos;
					}
					break;
				}
				++pos;
			}
			
			double est_freq = summary->estimate_frequency(itemset, len);
			for (int i=0; i<len; ++i)
				fprintf(estoutfile, "%i ", id2item[itemset[i]]);
			if (abssup != -1) {
				double log_pval = log_pvalue(trans_count, est_freq, abssup);
				fprintf(estoutfile, "\t%i\t%.15f\t%.15f\n", abssup, est_freq, exp(log_pval));
			}
			else
				fprintf(estoutfile, "\t%.15f\n", est_freq);
		}
		
	}
	
	estinfile->close();
	fclose(estoutfile);
	
	delete [] itemset;
	itemset = 0;
	
	return;
}


double SummaryMiner::get_penalty(int k, int x) const
{
	switch (penalty_type) {
		case 1:
			return get_bic_penalty(k);
			break;
		case 2:
			return get_mdl_penalty_1(k);
			break;
		case 3:
			return get_mdl_penalty_2(k, x);
			break;
		default:
			return 0.0;
	}
}


double SummaryMiner::get_bic_penalty(int k) const
{
	if (k == -1)
		k = summary->all_itemsets_size;
	
	return k / 2.0 * log_2(trans_count);
}


double SummaryMiner::get_mdl_penalty_1(int k) const
{
	if (k == -1)
		k = summary->all_itemsets_size;
	
	return k * (item_count + 1 + log_2(trans_count+1)) + 1 ;	
}


double SummaryMiner::get_mdl_penalty_2(int k, int x) const
{
	if (k == -1)
		k = summary->all_itemsets_size;
	
	if (x == -1) {
		x = 0;
		for (uint i=0; i<summary->all_itemsets_size; ++i)
			x += summary->all_itemsets[i]->len;
	}
	
	return k * (log_2(trans_count+1) + item_count * log_2(1.0 + 1.0/item_count) + 1) + log_2(item_count) * x + 1 ;
}


void SummaryMiner::read_dictionary(string dictionaryfilename)
{
	if (dictionaryfilename.size() == 0) {
		dict = 0;
		return;
	}
	
	init_summary();
	
	// one item name per line, position must match original item name, starting from 0
	dict = new string[summary->item_cnt+2];
	ifstream* dictionaryfile = new ifstream(dictionaryfilename.c_str());
	int linenr = 0;
	while (!dictionaryfile->eof()) {
		getline(*dictionaryfile, dict[linenr]);
		++linenr;
	}
	dictionaryfile->close();
	
	return;
}


void SummaryMiner::print_itemset(int i) const
{
	printf("%f\t", summary->frequencies[i]);
	
	int len = summary->all_itemsets[i]->len;
	
	if (dict == 0) {
		for (int j=0; j<len; ++j)
			printf("%i ", id2item[summary->all_itemsets[i]->itemset[j]]);
	}
	else {
		for (int j=0; j<len; ++j)
			printf("%s ", dict[id2item[summary->all_itemsets[i]->itemset[j]]].c_str());
	}
	
	printf("\n");
	
	return;
}


void SummaryMiner::print_summary() const
{
	for (uint i=0; i<summary->all_itemsets_size; ++i) {
		print_itemset(i);
	}
	
	return;
}


void SummaryMiner::save_itemset(int i, FILE* outfile) const
{
	fprintf(outfile, "%f ", summary->frequencies[i]);
	
	int len = summary->all_itemsets[i]->len;
	if (dict == 0) {
		for (int j=0; j<len; ++j) {
			int id = summary->all_itemsets[i]->itemset[j];
			fprintf(outfile, "%i ", id2item[id]);
		}
	}
	else {
		for (int j=0; j<len; ++j) {
			int id = summary->all_itemsets[i]->itemset[j];
			fprintf(outfile, "%s ", dict[id2item[id]].c_str());
		}
	}
	
	fprintf(outfile, "\n");
	
	return;
}


void SummaryMiner::save_summary(string outfilename) const
{
	FILE *outfile = fopen(outfilename.c_str(), "w");
	
	fprintf(outfile, "# Itemset Summary of '%s'\n", datafilename.c_str());
	fprintf(outfile, "#   rows:\t%i\n", (int)trans_count);
	fprintf(outfile, "#   items:\t%i\n", item_count);
	fprintf(outfile, "#   itemsets:\t%i\n", summary->all_itemsets_size);
	fprintf(outfile, "#   singletons:\t%s\n", (use_items)?"yes":"no");
	fprintf(outfile, "#   minsup:\t%f\n", minsup);
	fprintf(outfile, "#   maxsize:\t%i\n", maxsize);
	fprintf(outfile, "#   maxtreewidth:\t%i\n", maxtreewidth);
	fprintf(outfile, "#   maxgroupsize:\t%i\n", maxgroupsize);
	fprintf(outfile, "#   entropy:\t%f\n", summary->get_entropy());
	fprintf(outfile, "#\n");
	
	for (uint i=0; i<summary->all_itemsets_size; ++i) {
		save_itemset(i, outfile);
	}
	
	fclose(outfile);
	
	return;
}


void SummaryMiner::insert_top_k_itemset(Itemset & iset, double score, double freq, double est)
{
	int loc = top_k - 1;
	while (loc >= 0 && score > top_k_scores[loc]) --loc;
	++loc;
	
	for (int i=top_k-1; i>loc; --i) {
		top_k_itemsets[i] = top_k_itemsets[i-1];
		top_k_scores[i]   = top_k_scores[i-1];
		top_k_freqs[i]    = top_k_freqs[i-1];
		top_k_ests[i]     = top_k_ests[i-1];
	}
	
	top_k_itemsets[loc] = iset;
	top_k_scores[loc]   = score;
	top_k_freqs[loc]    = freq;
	top_k_ests[loc]     = est;
	
	return;
}


void SummaryMiner::print_top_k() const
{
	printf("Current top-%i itemsets\n", top_k);
	for (int i=0; i<top_k; ++i) {
		// sort the itemset prior to outputting, done here rather than during insertion
		sort(top_k_itemsets[i].itemset, top_k_itemsets[i].itemset + top_k_itemsets[i].len);
		
		printf("%f\t", top_k_freqs[i]);
		
		if (dict == 0) {
			for (int j=0; j<top_k_itemsets[i].len; ++j)
				printf("%i ", id2item[top_k_itemsets[i].itemset[j]]);
		}
		else {
			for (int j=0; j<top_k_itemsets[i].len; ++j)
				printf("%s ", dict[id2item[top_k_itemsets[i].itemset[j]]].c_str());
		}
		
		printf("\n");
	}
	
	printf("\n");
	
	return;
}
