/*
 * Template Trie structure
 * Author:
 */

#include "intmap.h"
#include <algorithm>
#include <string.h>


using namespace std;



template <class T> class Trie;


template <class T> class Node
{
public:
	Node(T val);
	~Node();

public:
	T val;
	IntMap<Node<T>* > children;
	friend class Trie<T>;
};



template <class T> class Trie
{
public:
	Trie(T defaultval = T());
	~Trie();
	void construct();

	int size() const { return trie_size; }
	Node<T> *add(const int *key, int keylen, T value, Node<T> *start=0);
	T find(const int *key, int keylen, Node<T> *start=0) const;
	Node<T>* locate(const int *key, int keylen, Node<T> *start=0) const;
	
	Node<T> * get_root() { return root; }
	
	int get_trie_size() const { return trie_size; }
	
private:
	void sort_key(int *&key, int keylen) const;
	void unsort_key(int *&key) const;

	int trie_size;
	Node<T> *root;

	const T defaultval;
	T notfound;

	static const bool sort_keys;
	static int *tmp_sort_key;
};

template <class T> const bool Trie<T>::sort_keys = true;
template <class T> int *Trie<T>::tmp_sort_key = 0;



template <class T> inline Node<T>::Node(T val): val(val)
{

}


template <class T> inline Node<T>::~Node()
{
	for (int i=0; i<children.size; ++i) {
		delete children.map[i].second;
	}
}


template <class T> Trie<T>::Trie(T defaultval) : trie_size(0), defaultval(defaultval)
{
	construct();
}


template <class T> inline void Trie<T>::construct()
{
	notfound = -1;
	root = new Node<T>(defaultval);
	if (tmp_sort_key==0) {
		const int MAX_KEY_SIZE = 10000; //
		tmp_sort_key = new int[MAX_KEY_SIZE];
	}
}


template <class T> Trie<T>::~Trie()
{
	delete root;
	delete [] tmp_sort_key;
	tmp_sort_key = 0;
}



template <class T> T Trie<T>::find(const int *key, int keylen, Node<T> *start) const
{
	Node<T> *loc = locate(key, keylen, start);
	if (loc)
		return loc->val;
	else
		return notfound;
}


template <class T> inline void Trie<T>::sort_key(int *&key, int keylen) const
{
	if (sort_keys) {
		memcpy(tmp_sort_key, key, keylen*sizeof(int));
		sort(tmp_sort_key, tmp_sort_key+keylen);
		swap(key, tmp_sort_key);
	}
}


template <class T> inline void Trie<T>::unsort_key(int *&key) const
{
	if (sort_keys) {
		swap(key, tmp_sort_key);
	}
}


template <class T> inline Node<T> *Trie<T>::locate(const int *key, int keylen, Node<T>* start) const
{
	int *thekey = (int*)key;
	Node<T> *current;
	if (start)
		current = start;
	else
		current = root;
	if (keylen==0)
		return current;
	
	sort_key(thekey, keylen);
	
	Node<T>* pit;
	int *sk = thekey;
	thekey += keylen;
	
	while (sk!=thekey) {
		pit = current->children.get(*sk);
		if (pit==0) {
			thekey -= keylen;
			unsort_key(thekey);
			return 0;
		}
		else
			current = pit;
		++sk;
	}
	
	thekey -= keylen;
	unsort_key(thekey);
	
	return current;
}


template <class T> Node<T>* Trie<T>::add(const int *key, int keylen, T value, Node<T> *start)
{
	int *thekey = (int*) key;
	sort_key(thekey, keylen);

	Node<T> *current = start;
	if (current==0)
		current = root;

	Node<T>* pit;
	int *sk = thekey;
	thekey += keylen;
	while (sk!=thekey) {
		pit = current->children.get(*sk);
		if (pit==0) {
			Node<T> *newnode = new Node<T>(defaultval);
			current->children.insert(*sk, newnode);
			current = newnode;
		}
		else
			current = pit;
		++sk;
	}

	current->val = value;
	++trie_size;

	thekey -= keylen;
	unsort_key(thekey);

	return current;
}
